import time

from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .receipt import ReceiptProtocolEntity
class OutgoingReceiptProtocolEntity(ReceiptProtocolEntity):

    '''
    delivered:
    If we send the following without "to" specified, whatsapp will consider the message delivered,
    but will not notify the sender.
    <receipt to="xxxxxxxxxxx@s.whatsapp.net" id="1415389947-15"></receipt>

    read
    <receipt to="xxxxxxxxxxx@s.whatsapp.net" id="1415389947-15" type="read"></receipt>

    multiple items:
    <receipt type="read" to="xxxxxxxxxxxx@s.whatsapp.net" id="1431364583-191">
        <list>
            <item id="1431364572-189"></item>
            <item id="1431364575-190"></item>
        </list>
    </receipt>
    '''


    def __init__(self, messageIds, to, read = False, participant = None, callId = None):
        if type(messageIds) in (list, tuple):
            if len(messageIds) > 1:
                receiptId = self._generateId()
            else:
                receiptId = messageIds[0]
        else:
            receiptId = messageIds
            messageIds = [messageIds]

        super(OutgoingReceiptProtocolEntity, self).__init__(receiptId)
        self.setOutgoingData(messageIds, to, read, participant, callId)

    def setOutgoingData(self, messageIds, to, read, participant, callId):
        self.messageIds = messageIds
        self.to = to
        self.read = read
        self.participant = participant
        self.callId = callId

    def getMessageIds(self):
        return self.messageIds

    def toProtocolTreeNode(self):
        node = super(OutgoingReceiptProtocolEntity, self).toProtocolTreeNode()
        if self.read:
            node.setAttribute("type", "read")
        if self.participant:
            node.setAttribute("participant", self.participant)
        if self.callId:
            offer = ProtocolTreeNode("offer", {"call-id": self.callId})
            node.addChild(offer)

        node.setAttribute("to", self.to)

        if len(self.messageIds) > 1:
            listNode = ProtocolTreeNode("list")
            listNode.addChildren([ProtocolTreeNode("item", {"id": mId}) for mId in self.messageIds])
            node.addChild(listNode)

        return node

    def __str__(self):
        out = super(OutgoingReceiptProtocolEntity, self).__str__()
        out  += "To: \n%s" % self.to
        if self.read:
            out += "Type: \n%s" % "read"
        out += "For: \n%s" % self.messageIds
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        listNode = node.getChild("list")
        messageIds = []
        if listNode:
            messageIds = [child["id"] for child in listNode.getChildren()]
        else:
            messageIds = [node["id"]]

        return OutgoingReceiptProtocolEntity(
            messageIds,
            node["to"],
            node["type"] == "read",
            node["participant"]
            )



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class AckProtocolEntity(ProtocolEntity):

    '''
    <ack class="{{receipt | message | ?}}" id="{{message_id}}">
    </ack>
    '''

    def __init__(self, _id, _class):
        super(AckProtocolEntity, self).__init__("ack")
        self._id = _id
        self._class = _class

    def getId(self):
        return self._id

    def getClass(self):
        return self._class
    
    def toProtocolTreeNode(self):
        attribs = {
            "id"           : self._id,
            "class"        : self._class,
        }

        return self._createProtocolTreeNode(attribs, None, data = None)

    def __str__(self):
        out  = "ACK:\n"
        out += "ID: %s\n" % self._id
        out += "Class: %s\n" % self._class
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        return AckProtocolEntity(
            node.getAttributeValue("id"),
            node.getAttributeValue("class")
            )



from yowsup.layers import YowLayer, YowLayerEvent, EventCallback
from yowsup.common.http.httpproxy import HttpProxy
from yowsup.layers.network.layer_interface import YowNetworkLayerInterface
import asyncore, socket, logging
logger = logging.getLogger(__name__)

class YowNetworkLayer(YowLayer, asyncore.dispatcher_with_send):
    '''
        send:       bytearray -> None
        receive:    bytearray -> bytearray
    '''

    EVENT_STATE_CONNECT         = "org.openwhatsapp.yowsup.event.network.connect"
    EVENT_STATE_DISCONNECT      = "org.openwhatsapp.yowsup.event.network.disconnect"
    EVENT_STATE_CONNECTED       = "org.openwhatsapp.yowsup.event.network.connected"
    EVENT_STATE_DISCONNECTED    = "org.openwhatsapp.yowsup.event.network.disconnected"

    PROP_ENDPOINT               = "org.openwhatsapp.yowsup.prop.endpoint"
    PROP_NET_READSIZE           = "org.openwhatsapp.yowsup.prop.net.readSize"

    def __init__(self):
        asyncore.dispatcher.__init__(self)
        YowLayer.__init__(self)
        self.interface = YowNetworkLayerInterface(self)
        httpProxy = HttpProxy.getFromEnviron()
        proxyHandler = None
        if httpProxy != None:
            logger.debug("HttpProxy initialize: %s" % httpProxy)
            def onConnect():
                logger.debug("HttpProxy connected")
                self.proxyHandler = None
                self.handle_connect()
            proxyHandler = httpProxy.handler()
            proxyHandler.onConnect = onConnect
        self.proxyHandler = proxyHandler
        
    @EventCallback(EVENT_STATE_CONNECT)
    def onConnect(self, ev):
        self.createConnection()
        return True
    
    @EventCallback(EVENT_STATE_DISCONNECT)
    def onDisconnect(self, ev):
        self.destroyConnection(ev.getArg("reason"))
        return True        

    def createConnection(self):
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.out_buffer = bytearray()
        endpoint = self.getProp(self.__class__.PROP_ENDPOINT)
        logger.debug("Connecting to %s:%s" % endpoint)
        if self.proxyHandler != None:
            logger.debug("HttpProxy connect: %s:%d" % endpoint)
            self.proxyHandler.connect(self, endpoint)
        else:
            self.connect(endpoint)

    def destroyConnection(self, reason = None):
        self.handle_close(reason or "Requested")

    def getStatus(self):
        return self.connected

    def handle_connect(self):
        self.connected = True
        if self.proxyHandler != None:
            logger.debug("HttpProxy handle connect")
            self.proxyHandler.send(self)
        else:
            self.emitEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECTED))

    def handle_close(self, reason = "Connection Closed"):
        self.connected = False
        logger.debug("Disconnected, reason: %s" % reason)
        self.emitEvent(YowLayerEvent(self.__class__.EVENT_STATE_DISCONNECTED, reason = reason, detached=True))
        self.close()

    def handle_error(self):
        raise

    def handle_read(self):
        readSize = self.getProp(self.__class__.PROP_NET_READSIZE, 1024)
        if self.proxyHandler != None:
            data = self.proxyHandler.recv(self, readSize)
            logger.debug("HttpProxy handle read: %s" % data)
        else:
            data = self.recv(readSize)
            self.receive(data)

    def send(self, data):
        if self.connected:
            self.out_buffer = self.out_buffer + data
            self.initiate_send()

    def receive(self, data):
        self.toUpper(data)

    def __str__(self):
        return "Network Layer"



from yowsup.layers.protocol_presence.protocolentities.presence import PresenceProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class PresenceProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = PresenceProtocolEntity
        self.node = ProtocolTreeNode("presence", {"type": "presence_type", "name": "presence_name"}, None, None)



from yowsup.layers.protocol_presence.protocolentities.presence_subscribe import SubscribePresenceProtocolEntity
from yowsup.layers.protocol_presence.protocolentities.test_presence import PresenceProtocolEntityTest

class SubscribePresenceProtocolEntityTest(PresenceProtocolEntityTest):
    def setUp(self):
        super(SubscribePresenceProtocolEntityTest, self).setUp()
        self.ProtocolEntity = SubscribePresenceProtocolEntity
        self.node.setAttribute("type", "subscribe")
        self.node.setAttribute("to", "subscribe_jid")


from .yowstack import YowStack, YowStackBuilder

from yowsup.layers.auth                        import YowCryptLayer, YowAuthenticationProtocolLayer, AuthError
from yowsup.layers.coder                       import YowCoderLayer
from yowsup.layers.logger                      import YowLoggerLayer
from yowsup.layers.network                     import YowNetworkLayer
from yowsup.layers.protocol_messages           import YowMessagesProtocolLayer
from yowsup.layers.stanzaregulator             import YowStanzaRegulator
from yowsup.layers.protocol_media              import YowMediaProtocolLayer
from yowsup.layers.protocol_acks               import YowAckProtocolLayer
from yowsup.layers.protocol_receipts           import YowReceiptProtocolLayer
from yowsup.layers.protocol_groups             import YowGroupsProtocolLayer
from yowsup.layers.protocol_presence           import YowPresenceProtocolLayer
from yowsup.layers.protocol_ib                 import YowIbProtocolLayer
from yowsup.layers.protocol_notifications      import YowNotificationsProtocolLayer
from yowsup.layers.protocol_iq                 import YowIqProtocolLayer
from yowsup.layers.protocol_contacts           import YowContactsIqProtocolLayer
from yowsup.layers.protocol_chatstate          import YowChatstateProtocolLayer
from yowsup.layers.protocol_privacy            import YowPrivacyProtocolLayer
from yowsup.layers.protocol_profiles           import YowProfilesProtocolLayer
from yowsup.layers.protocol_calls              import YowCallsProtocolLayer



YOWSUP_CORE_LAYERS = (
    YowLoggerLayer,
    YowCoderLayer,
    YowCryptLayer,
    YowStanzaRegulator,
    YowNetworkLayer
)


YOWSUP_PROTOCOL_LAYERS_BASIC = (
    YowAuthenticationProtocolLayer, YowMessagesProtocolLayer,
    YowReceiptProtocolLayer, YowAckProtocolLayer, YowPresenceProtocolLayer,
    YowIbProtocolLayer, YowIqProtocolLayer, YowNotificationsProtocolLayer,
    YowContactsIqProtocolLayer, YowChatstateProtocolLayer

)

YOWSUP_PROTOCOL_LAYERS_GROUPS = (YowGroupsProtocolLayer,) + YOWSUP_PROTOCOL_LAYERS_BASIC
YOWSUP_PROTOCOL_LAYERS_MEDIA  = (YowMediaProtocolLayer,) + YOWSUP_PROTOCOL_LAYERS_BASIC
YOWSUP_PROTOCOL_LAYERS_PROFILES  = (YowProfilesProtocolLayer,) + YOWSUP_PROTOCOL_LAYERS_BASIC
YOWSUP_PROTOCOL_LAYERS_CALLS  = (YowCallsProtocolLayer,) + YOWSUP_PROTOCOL_LAYERS_BASIC
YOWSUP_PROTOCOL_LAYERS_FULL = (YowGroupsProtocolLayer, YowMediaProtocolLayer, YowPrivacyProtocolLayer, YowProfilesProtocolLayer, YowCallsProtocolLayer)\
                              + YOWSUP_PROTOCOL_LAYERS_BASIC


YOWSUP_FULL_STACK = (YOWSUP_PROTOCOL_LAYERS_FULL,) +\
                     YOWSUP_CORE_LAYERS



from yowsup.layers.protocol_media.protocolentities.message_media_downloadable_video import VideoDownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.test_message_media_downloadable import DownloadableMediaMessageProtocolEntityTest

class VideoDownloadableMediaMessageProtocolEntityTest(DownloadableMediaMessageProtocolEntityTest):
    def setUp(self):
        super(VideoDownloadableMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = VideoDownloadableMediaMessageProtocolEntity
        mediaNode = self.node.getChild("media")
        mediaNode.setAttribute("abitrate",  "165")
        mediaNode.setAttribute("acodec",    "aac")
        mediaNode.setAttribute("asampfmt",  "flt")
        mediaNode.setAttribute("asampfreq", "48000")
        mediaNode.setAttribute("duration",  "61")
        mediaNode.setAttribute("encoding",  "raw")
        mediaNode.setAttribute("fps",       "24")
        mediaNode.setAttribute("height",    "452")
        mediaNode.setAttribute("seconds",   "61")
        mediaNode.setAttribute("vbitrate",  "1862")
        mediaNode.setAttribute("vcodec",    "h264")
        mediaNode.setAttribute("width",     "800")



from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from .protocolentities import * 
class YowChatstateProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "chatstate": (self.recvChatstateNode, self.sendChatstateEntity)
        }
        super(YowChatstateProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Chatstate Layer"

    def sendChatstateEntity(self, entity):
        self.entityToLower(entity)

    def recvChatstateNode(self, node):
        self.toUpper(IncomingChatstateProtocolEntity.fromProtocolTreeNode(node))



from yowsup.common import YowConstants
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_iq.protocolentities import ResultIqProtocolEntity
class SuccessCreateGroupsIqProtocolEntity(ResultIqProtocolEntity):
    '''
    <iq type="result" id="{{id}}" from="g.us">
        <group id="{group_id}"></group>
    </iq>
    '''

    def __init__(self, _id, groupId):
        super(SuccessCreateGroupsIqProtocolEntity, self).__init__(_from = YowConstants.WHATSAPP_GROUP_SERVER, _id = _id)
        self.setProps(groupId)

    def setProps(self, groupId):
        self.groupId = groupId

    def toProtocolTreeNode(self):
        node = super(SuccessCreateGroupsIqProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode("group",{"id": self.groupId}))
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(SuccessCreateGroupsIqProtocolEntity, SuccessCreateGroupsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = SuccessCreateGroupsIqProtocolEntity
        entity.setProps(node.getChild("group").getAttributeValue("id"))
        return entity



from yowsup.layers.protocol_receipts.protocolentities import OutgoingReceiptProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class OutgoingReceiptProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = OutgoingReceiptProtocolEntity
        self.node = OutgoingReceiptProtocolEntity("123", "target", "read").toProtocolTreeNode()


import unittest
import inspect
try:
    import Queue
except ImportError:
    import queue as Queue

class YowLayerEvent:
    def __init__(self, name, **kwargs):
        self.name = name
        self.detached = False
        if "detached" in kwargs:
            del kwargs["detached"]
            self.detached = True
        self.args = kwargs

    def isDetached(self):
        return self.detached

    def getName(self):
        return self.name

    def getArg(self, name):
        return self.args[name] if name in self.args else None
    
class EventCallback(object):
    def __init__(self, eventName):
        self.eventName = eventName

    def __call__(self, fn):
        fn.event_callback = self.eventName
        return fn


class YowLayer(object):
    __upper = None
    __lower = None
    _props = {}
    __detachedQueue = Queue.Queue()
    # def __init__(self, upperLayer, lowerLayer):
    #     self.setLayers(upperLayer, lowerLayer)

    def __init__(self):
        self.setLayers(None, None)
        self.interface = None
        self.event_callbacks = {}
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        for m in members:
            if hasattr(m[1], "event_callback"):
                fname = m[0]
                fn = m[1]
                self.event_callbacks[fn.event_callback] = getattr(self, fname)

    def getLayerInterface(self, YowLayerClass = None):
        return self.interface if YowLayerClass is None else self.__stack.getLayerInterface(YowLayerClass)

    def setStack(self, stack):
        self.__stack = stack

    def getStack(self):
        return self.__stack

    def setLayers(self, upper, lower):
        self.__upper = upper
        self.__lower = lower

    def send(self, data):
        self.toLower(data)

    def receive(self, data):
        self.toUpper(data)

    def toUpper(self, data):
        if self.__upper:
            self.__upper.receive(data)

    def toLower(self, data):
        if self.__lower:
            self.__lower.send(data)

    def emitEvent(self, yowLayerEvent):
        if self.__upper and not self.__upper.onEvent(yowLayerEvent):
            if yowLayerEvent.isDetached():
                yowLayerEvent.detached = False
                self.getStack().execDetached(lambda :  self.__upper.emitEvent(yowLayerEvent))

            else:
                self.__upper.emitEvent(yowLayerEvent)


    def broadcastEvent(self, yowLayerEvent):
        if self.__lower and not self.__lower.onEvent(yowLayerEvent):
            if yowLayerEvent.isDetached():
                yowLayerEvent.detached = False
                self.getStack().execDetached(lambda:self.__lower.broadcastEvent(yowLayerEvent))
            else:
                self.__lower.broadcastEvent(yowLayerEvent)

    '''return true to stop propagating the event'''
    def onEvent(self, yowLayerEvent):
        eventName = yowLayerEvent.getName()
        if eventName in self.event_callbacks:
            return self.event_callbacks[eventName](yowLayerEvent)
        return False

    def getProp(self, key, default = None):
        return self.getStack().getProp(key, default)

    def setProp(self, key, val):
        return self.getStack().setProp(key, val)


class YowProtocolLayer(YowLayer):
    def __init__(self, handleMap = None):
        super(YowProtocolLayer, self).__init__()
        self.handleMap = handleMap or {}
        self.iqRegistry = {}

    def receive(self, node):
        if not self.processIqRegistry(node):
            if node.tag in self.handleMap:
                recv, _ = self.handleMap[node.tag]
                if recv:
                    recv(node)

    def send(self, entity):
        if entity.getTag() in self.handleMap:
            _, send = self.handleMap[entity.getTag()]
            if send:
                send(entity)

    def entityToLower(self, entity):
        #super(YowProtocolLayer, self).toLower(entity.toProtocolTreeNode())
        self.toLower(entity.toProtocolTreeNode())

    def isGroupJid(self, jid):
        return "-" in jid

    def raiseErrorForNode(self, node):
        raise ValueError("Unimplemented notification type %s " % node)


    def _sendIq(self, iqEntity, onSuccess = None, onError = None):
        self.iqRegistry[iqEntity.getId()] = (iqEntity, onSuccess, onError)
        self.toLower(iqEntity.toProtocolTreeNode())

    def processIqRegistry(self, protocolTreeNode):
        if protocolTreeNode.tag == "iq":
            iq_id = protocolTreeNode["id"]
            if iq_id in self.iqRegistry:
                originalIq, successClbk, errorClbk = self.iqRegistry[iq_id]
                del self.iqRegistry[iq_id]

                if protocolTreeNode["type"] == "result" and successClbk:
                    successClbk(protocolTreeNode, originalIq)
                elif protocolTreeNode["type"] == "error" and errorClbk:
                    errorClbk(protocolTreeNode, originalIq)
                return True

        return False

class YowParallelLayer(YowLayer):
    def __init__(self, sublayers = None):
        super(YowParallelLayer, self).__init__()
        self.sublayers = sublayers or []
        self.sublayers = tuple([sublayer() for sublayer in sublayers])
        for s in self.sublayers:
            #s.setLayers(self, self)
            s.toLower = self.toLower
            s.toUpper = self.toUpper
            s.broadcastEvent = self.subBroadcastEvent
            s.emitEvent = self.subEmitEvent


    def getLayerInterface(self, YowLayerClass):
        for s in self.sublayers:
            if s.__class__ == YowLayerClass:
                return s.getLayerInterface()

    def setStack(self, stack):
        super(YowParallelLayer, self).setStack(stack)
        for s in self.sublayers:
            s.setStack(self.getStack())


    def receive(self, data):
        for s in self.sublayers:
            s.receive(data)

    def send(self, data):
        for s in self.sublayers:
            s.send(data)

    def subBroadcastEvent(self, yowLayerEvent):
        self.onEvent(yowLayerEvent)
        self.broadcastEvent(yowLayerEvent)

    def subEmitEvent(self, yowLayerEvent):
        self.onEvent(yowLayerEvent)
        self.emitEvent(yowLayerEvent)


    def onEvent(self, yowLayerEvent):
        stopEvent = False
        for s in self.sublayers:
            stopEvent = stopEvent or s.onEvent(yowLayerEvent)

        return stopEvent

    def __str__(self):
        return " - ".join([l.__str__() for l in self.sublayers])

class YowLayerInterface(object):
    def __init__(self, layer):
        self._layer = layer


class YowLayerTest(unittest.TestCase):
    def __init__(self, *args):
        super(YowLayerTest, self).__init__(*args)
        self.upperSink = []
        self.lowerSink = []
        self.toUpper = self.receiveOverrider
        self.toLower = self.sendOverrider
        self.upperEventSink = []
        self.lowerEventSink = []
        self.emitEvent = self.emitEventOverrider
        self.broadcastEvent = self.broadcastEventOverrider

    def receiveOverrider(self, data):
        self.upperSink.append(data)

    def sendOverrider(self, data):
        self.lowerSink.append(data)
        
    def emitEventOverrider(self, event):
        self.upperEventSink.append(event)
    
    def broadcastEventOverrider(self, event):
        self.lowerEventSink.append(event)
        
    def assert_emitEvent(self, event):
        self.emitEvent(event)
        try:
            self.assertEqual(event, self.upperEventSink.pop())
        except IndexError:
            raise AssertionError("Event '%s' was not emited through this layer" % (event.getName()))
        
    def assert_broadcastEvent(self, event):
        self.broadcastEvent(event)
        try:
            self.assertEqual(event, self.lowerEventSink.pop())
        except IndexError:
            raise AssertionError("Event '%s' was not broadcasted through this layer" % (event.getName()))

class YowProtocolLayerTest(YowLayerTest):
    def assertSent(self, entity):
        self.send(entity)
        try:
            self.assertEqual(entity.toProtocolTreeNode(), self.lowerSink.pop())
        except IndexError:
            raise AssertionError("Entity '%s' was not sent through this layer" % (entity.getTag()))

    def assertReceived(self, entity):
        node = entity.toProtocolTreeNode()
        self.receive(node)
        try:
            self.assertEqual(node, self.upperSink.pop().toProtocolTreeNode())
        except IndexError:
            raise AssertionError("'%s' was not received through this layer" % (entity.getTag()))



from .iq import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class PushIqProtocolEntity(IqProtocolEntity):
    def __init__(self):
        super(PushIqProtocolEntity, self).__init__("urn:xmpp:whatsapp:push", _type="get")

    def toProtocolTreeNode(self):
        node = super(PushIqProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode("config"))
        return node


from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode

class SetStatusIqProtocolEntity(IqProtocolEntity):
    '''
    <iq to="s.whatsapp.net" xmlns="status" type="set" id="{{IQ_ID}}">
        <status>{{MSG}}</status>
    </notification>
    '''
    XMLNS = "status"
    def __init__(self, text = None, _id = None):
        super(SetStatusIqProtocolEntity, self).__init__(self.__class__.XMLNS, _id, _type = "set", to = YowConstants.WHATSAPP_SERVER)
        self.setData(text)
        
    def setData(self, text):
        self.text = text

    def toProtocolTreeNode(self):
        node = super(SetStatusIqProtocolEntity, self).toProtocolTreeNode()
        statusNode = ProtocolTreeNode("status", {}, [], self.text)
        node.addChild(statusNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = SetStatusIqProtocolEntity
        statusNode = node.getChild("status")
        entity.setData(statusNode.getData())
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class IqProtocolEntity(ProtocolEntity):

    '''
    <iq type="{{get | set}}" id="{{id}}" xmlns="{{xmlns}}" to="{{TO}}" from="{{FROM}}">
    </iq>
    '''

    TYPE_SET = "set"
    TYPE_GET = "get"
    TYPE_ERROR = "error"
    TYPE_RESULT = "result"

    TYPES = (TYPE_SET, TYPE_GET, TYPE_RESULT, TYPE_ERROR)
    def __init__(self, xmlns = None, _id = None, _type = None, to = None, _from = None):
        super(IqProtocolEntity, self).__init__("iq")

        assert _type in self.__class__.TYPES, "Iq of type %s is not implemented, can accept only (%s)" % (_type," | ".join(self.__class__.TYPES))
        assert not to or not _from, "Can't set from and to at the same time"
        self._id = self._generateId(True) if _id is None else _id
        self._from = _from
        self._type = _type
        self.xmlns = xmlns
        self.to = to

    def getId(self):
        return self._id

    def getType(self):
        return self._type

    def getXmlns(self):
        return self.xmlns

    def getFrom(self, full = True):
        return self._from if full else self._from.split('@')[0]

    def getTo(self):
        return self.to
    
    def toProtocolTreeNode(self):
        attribs = {
            "id"          : self._id,
            "type"        : self._type
        }

        if self.xmlns:
            attribs["xmlns"] = self.xmlns

        if self.to:
            attribs["to"] = self.to
        elif self._from:
            attribs["from"] = self._from

        return self._createProtocolTreeNode(attribs, None, data = None)

    def __str__(self):
        out  = "Iq:\n"
        out += "ID: %s\n" % self._id
        out += "Type: %s\n" % self._type
        if self.xmlns:
            out += "xmlns: %s\n" % self.xmlns
        if self.to:
            out += "to: %s\n" % self.to
        elif self._from:
            out += "from: %s\n" % self._from
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        return IqProtocolEntity(
            node.getAttributeValue("xmlns"),
            node.getAttributeValue("id"),
            node.getAttributeValue("type"),
            node.getAttributeValue("to"),
            node.getAttributeValue("from")
            )



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class StreamErrorConflictProtocolEntity(ProtocolEntity):
    '''
     <stream:error>
        <conflict></conflict>
        <text>Replaced by new connection</text>
     </stream:error>
    '''
    def __init__(self,  text = None):
        super(StreamErrorConflictProtocolEntity, self).__init__("stream:error")
        self.setText(text)

    def setText(self, text = None):
        self.text = text or ''

    def getText(self):
        return self.text

    def __str__(self):
        out  = "Conflict Stream Error\n"
        if self.text:
            out += "Text: %s\n" % self.getText()

        return out

    def toProtocolTreeNode(self):
        node = super(StreamErrorConflictProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode("conflict"))
        node.addChild(ProtocolTreeNode("text", data=self.text))
        return node


    @staticmethod
    def fromProtocolTreeNode(node):
        return StreamErrorConflictProtocolEntity(node.getChild("text").getData())



from yowsup.layers.protocol_notifications.protocolentities.notification_picture import PictureNotificationProtocolEntity
from yowsup.layers.protocol_notifications.protocolentities.test_notification import NotificationProtocolEntityTest

class PictureNotificationProtocolEntityTest(NotificationProtocolEntityTest):
    def setUp(self):
        super(PictureNotificationProtocolEntityTest, self).setUp()
        self.ProtocolEntity = PictureNotificationProtocolEntity



from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class GetKeysIqProtocolEntity(IqProtocolEntity):
    def __init__(self, jids):
        super(GetKeysIqProtocolEntity, self).__init__("encrypt", _type = "get", to = YowConstants.WHATSAPP_SERVER)
        self.setJids(jids)

    def setJids(self, jids):
        assert type(jids) is list, "expected list of jids, got %s" % type(jids)
        self.jids = jids

    def getJids(self):
        return self.jids

    def toProtocolTreeNode(self):
        node = super(GetKeysIqProtocolEntity, self).toProtocolTreeNode()
        keyNode = ProtocolTreeNode("key")

        for jid in self.getJids():
            userNode = ProtocolTreeNode("user", { "jid": jid })
            keyNode.addChild(userNode)
        node.addChild(keyNode)
        return node



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_groups import GroupsIqProtocolEntity
class InfoGroupsIqProtocolEntity(GroupsIqProtocolEntity):
    '''
    <iq id="{{id}}"" type="get" to="{{group_jid}}" xmlns="w:g2">
        <query request="interactive"></query>
    </iq>


    '''

    def __init__(self, group_jid, _id=None):
        super(InfoGroupsIqProtocolEntity, self).__init__(to = group_jid, _id = _id, _type = "get")
        self.setProps(group_jid)

    def setProps(self, group_jid):
        self.group_jid = group_jid

    def __str__(self):
        out = super(InfoGroupsIqProtocolEntity, self).__str__()
        out += "Group JID: %s\n" % self.group_jid
        return out

    def toProtocolTreeNode(self):
        node = super(InfoGroupsIqProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode("query", {"request": "interactive"}))
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        assert node.getChild("query") is not None, "Not a groups info iq node %s" % node
        entity = super(InfoGroupsIqProtocolEntity, InfoGroupsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = InfoGroupsIqProtocolEntity
        entity.setProps(node.getAttributeValue("to"))
        return entity



from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities import ResultIqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class ResultRequestUploadIqProtocolEntity(ResultIqProtocolEntity):
    def __init__(self, _id, url, ip = None, resumeOffset = 0, duplicate = False):
        super(ResultRequestUploadIqProtocolEntity, self).__init__(_id = _id, _from = YowConstants.WHATSAPP_SERVER)
        self.setUploadProps(url, ip, resumeOffset, duplicate)

    def setUploadProps(self, url ,ip = None, resumeOffset = 0, duplicate = False):
        self.url = url
        self.ip = ip
        self.resumeOffset = resumeOffset or 0
        self.duplicate = duplicate

    def isDuplicate(self):
        return self.duplicate

    def getUrl(self):
        return self.url

    def getResumeOffset(self):
        return self.resumeOffset

    def getIp(self):
        return self.ip

    def __str__(self):
        out = super(ResultRequestUploadIqProtocolEntity, self).__str__()
        out += "URL: %s\n" % self.url
        if self.ip:
            out += "IP: %s\n" % self.ip
        return out

    def toProtocolTreeNode(self):
        node = super(ResultRequestUploadIqProtocolEntity, self).toProtocolTreeNode()

        if not self.isDuplicate():
            mediaNode = ProtocolTreeNode("media", {"url": self.url})
            if self.ip:
                mediaNode["ip"] = self.ip

            if self.resumeOffset:
                mediaNode["resume"] = str(self.resumeOffset)
        else:
            mediaNode = ProtocolTreeNode("duplicate", {"url": self.url})

        node.addChild(mediaNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity= ResultIqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ResultRequestUploadIqProtocolEntity
        mediaNode = node.getChild("media")
        if mediaNode:
            entity.setUploadProps(mediaNode["url"], mediaNode["ip"], mediaNode["resume"])
        else:
            duplicateNode = node.getChild("duplicate")
            if duplicateNode:
                entity.setUploadProps(duplicateNode["url"], duplicateNode["ip"], duplicate = True)
        return entity



from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest
from yowsup.layers.protocol_profiles.protocolentities import SetPrivacyIqProtocolEntity
from yowsup.structs import ProtocolTreeNode

entity = SetPrivacyIqProtocolEntity("all", ["profile","last","status"])

class SetPrivacyIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(SetPrivacyIqProtocolEntityTest, self).setUp()
        self.ProtocolEntity = SetPrivacyIqProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.common.http.warequest import WARequest
from yowsup.common.http.waresponseparser import JSONResponseParser
import socket, ssl, os, hashlib, sys
from time import sleep
import threading
import logging
from yowsup.common.tools import MimeTools

logger = logging.getLogger(__name__)

class MediaUploader(WARequest, threading.Thread):
    def __init__(self, jid, accountJid, sourcePath, uploadUrl, resumeOffset = 0, successClbk = None, errorClbk = None, progressCallback = None, async = True):
        WARequest.__init__(self)

        self.async = async
        self.jid = jid
        self.accountJid = accountJid
        self.sourcePath = sourcePath
        self.uploadUrl = uploadUrl
        self.resumeOffset = resumeOffset

        self.successCallback = successClbk
        self.errorCallback = errorClbk
        self.progressCallback = progressCallback

        self.pvars = ["name", "type", "size", "url", "error", "mimetype", "filehash", "width", "height"]

        self.setParser(JSONResponseParser())

        self.sock = socket.socket()

    def start(self):
        if self.async:
            threading.Thread.__init__(self)
            super(MediaUploader, self).start()
        else:
            self.run()

    def run(self):

        sourcePath = self.sourcePath
        uploadUrl = self.uploadUrl
        _host = uploadUrl.replace("https://","")

        self.url = _host[:_host.index('/')]


        try:
            filename = os.path.basename(sourcePath)
            filetype = MimeTools.getMIME(filename)
            filesize = os.path.getsize(sourcePath)

            self.sock.connect((self.url, self.port))
            ssl_sock = ssl.wrap_socket(self.sock)

            m = hashlib.md5()
            m.update(filename.encode())
            crypto = m.hexdigest() + os.path.splitext(filename)[1]

            boundary = "zzXXzzYYzzXXzzQQ"#"-------" + m.hexdigest() #"zzXXzzYYzzXXzzQQ"
            contentLength = 0

            hBAOS = "--" + boundary + "\r\n"
            hBAOS += "Content-Disposition: form-data; name=\"to\"\r\n\r\n"
            hBAOS += self.jid + "\r\n"
            hBAOS += "--" + boundary + "\r\n"
            hBAOS += "Content-Disposition: form-data; name=\"from\"\r\n\r\n"
            hBAOS += self.accountJid.replace("@whatsapp.net","") + "\r\n"

            hBAOS += "--" + boundary + "\r\n"
            hBAOS += "Content-Disposition: form-data; name=\"file\"; filename=\"" + crypto + "\"\r\n"
            hBAOS  += "Content-Type: " + filetype + "\r\n\r\n"

            fBAOS = "\r\n--" + boundary + "--\r\n"

            contentLength += len(hBAOS)
            contentLength += len(fBAOS)
            contentLength += filesize

            POST = "POST %s\r\n" % uploadUrl
            POST += "Content-Type: multipart/form-data; boundary=" + boundary + "\r\n"
            POST += "Host: %s\r\n" % self.url
            POST += "User-Agent: %s\r\n" % self.getUserAgent()
            POST += "Content-Length: " + str(contentLength) + "\r\n\r\n"

            ssl_sock.write(bytearray(POST.encode()))
            ssl_sock.write(bytearray(hBAOS.encode()))

            totalsent = 0
            buf = 1024
            f = open(sourcePath, 'rb')
            stream = f.read()
            f.close()
            status = 0
            lastEmit = 0

            while totalsent < int(filesize):
                ssl_sock.write(stream[:buf])
                status = totalsent * 100 / filesize
                if lastEmit!=status and status!=100 and filesize>12288:
                    if self.progressCallback:
                        self.progressCallback(self.sourcePath, self.jid, uploadUrl, int(status))
                lastEmit = status
                stream = stream[buf:]
                totalsent = totalsent + buf

            ssl_sock.write(bytearray(fBAOS.encode()))

            sleep(1)
            data = ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)

            if self.progressCallback:
                self.progressCallback(self.sourcePath, self.jid, uploadUrl, 100)


            lines = data.decode().splitlines()


            result = None

            for l in lines:
                if l.startswith("{"):
                    result = self.parser.parse(l, self.pvars)
                    break

            if not result:
                raise Exception("json data not found")


            if result["url"] is not None:
                if self.successCallback:
                    self.successCallback(sourcePath, self.jid, result["url"])
            else:
                logger.exception("uploadUrl: %s, result of uploading media has no url" % uploadUrl)
                if self.errorCallback:
                    self.errorCallback(sourcePath, self.jid, uploadUrl)

        except:
            logger.exception("Error occured at transfer %s"%sys.exc_info()[1])
            if self.errorCallback:
                self.errorCallback(sourcePath, self.jid, uploadUrl)



from yowsup.layers.protocol_media.protocolentities.test_message_media import MediaMessageProtocolEntityTest
from yowsup.layers.protocol_media.protocolentities import LocationMediaMessageProtocolEntity
from yowsup.structs import ProtocolTreeNode
class LocationMediaMessageProtocolEntityTest(MediaMessageProtocolEntityTest):
    def setUp(self):
        super(LocationMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = LocationMediaMessageProtocolEntity
        mediaNode = self.node.getChild("media")
        mediaNode["type"] = "location"
        mediaNode["latitude"] = "52.52393"
        mediaNode["longitude"] = "13.41747"
        mediaNode["encoding"] = "raw"



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .ib import IbProtocolEntity
class AccountIbProtocolEntity(IbProtocolEntity):
    '''
    <ib from="s.whatsapp.net">
        <account status="active | ?" kind="paid" creation="timestamp" expiration="timestamp"></account>
    </ib>
    '''

    STATUS_ACTIVE = "active"
    KIND_PAD      = "paid"

    def __init__(self, status, kind, creation, expiration):
        super(AccountIbProtocolEntity, self).__init__()
        self.setProps(status, kind, creation, expiration)


    def setProps(self, status, kind, creation, expiration):
        self.status = status
        self.creation = int(creation)
        self.kind = kind
        self.expiration= int(expiration)

    def toProtocolTreeNode(self):
        node = super(AccountIbProtocolEntity, self).toProtocolTreeNode()
        accountChild = ProtocolTreeNode("account",
                                        {
                                            "status": self.status,
                                            "kind": self.kind,
                                            "creation": int(self.creation),
                                            "expiration": int(self.expiration)

                                        })
        node.addChild(accountChild)
        return node

    def __str__(self):
        out = super(AccountIbProtocolEntity, self).__str__()
        out += "Status: %s\n" % self.status
        out += "Kind: %s\n" % self.kind
        out += "Creation: %s\n" % self.creation
        out += "Expiration: %s\n" % self.expiration

        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IbProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = AccountIbProtocolEntity
        accountNode = node.getChild("account")
        entity.setProps(
            accountNode["status"],
            accountNode["kind"],
            accountNode["creation"],
            accountNode["expiration"]
        )


from yowsup.layers.protocol_ib.protocolentities.test_ib import IbProtocolEntityTest
from yowsup.layers.protocol_ib.protocolentities.offline_ib import OfflineIbProtocolEntity
from yowsup.structs import ProtocolTreeNode
class OfflineIbProtocolEntityTest(IbProtocolEntityTest):
    def setUp(self):
        super(OfflineIbProtocolEntityTest, self).setUp()
        self.ProtocolEntity = OfflineIbProtocolEntity
        self.node.addChild(ProtocolTreeNode("offline", {"count": "5"}))


from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest
from yowsup.layers.protocol_profiles.protocolentities import ResultPrivacyIqProtocolEntity
from yowsup.structs import ProtocolTreeNode

entity = ResultPrivacyIqProtocolEntity({"profile":"all","last":"none","status":"contacts"})

class ResultPrivacyIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(ResultPrivacyIqProtocolEntityTest, self).setUp()
        self.ProtocolEntity = ResultPrivacyIqProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.layers.protocol_groups.protocolentities.iq_groups_create import CreateGroupsIqProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

entity = CreateGroupsIqProtocolEntity("group subject")

class CreateGroupsIqProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = CreateGroupsIqProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.layers import YowProtocolLayerTest
from yowsup.layers.protocol_contacts import YowContactsIqProtocolLayer
from yowsup.layers.protocol_contacts.protocolentities.test_notification_contact_add import entity as addEntity
from yowsup.layers.protocol_contacts.protocolentities.test_notification_contact_update import entity as updateEntity
from yowsup.layers.protocol_contacts.protocolentities.test_notification_contact_remove import entity as removeEntity
from yowsup.layers.protocol_contacts.protocolentities.test_iq_sync_result import entity as syncResultEntity
from yowsup.layers.protocol_contacts.protocolentities.test_iq_sync_get import entity as syncGetEntity

class YowContactsIqProtocolLayerTest(YowProtocolLayerTest, YowContactsIqProtocolLayer):
    def setUp(self):
        YowContactsIqProtocolLayer.__init__(self)

    def test_sync(self):
        self.assertSent(syncGetEntity)

    def test_syncResult(self):
        self.assertReceived(syncResultEntity)

    def test_notificationAdd(self):
        self.assertReceived(addEntity)

    def test_notificationUpdate(self):
        self.assertReceived(updateEntity)

    def test_notificationRemove(self):
        self.assertReceived(removeEntity)



import sys, tempfile, logging
logger = logging.getLogger(__name__)

if sys.version_info >= (3, 0):
    from urllib.request import urlopen
    from urllib.parse import urlencode
else:
    from urllib2 import urlopen
    from urllib import urlencode


class MediaDownloader:
    def __init__(self, successClbk = None, errorClbk = None, progressCallback = None):
        self.successCallback = successClbk
        self.errorCallback = errorClbk
        self.progressCallback = progressCallback

    def download(self, url = ""):
        try:
            
            if not url:
                if self.url:
                    url = "https://" if self.port == 443 else "http://"
                    url = url + self.url
                    url = url + "?" + urlencode(self.params)
                    logger.debug("URL is %s" % url)
                else:
                    raise Exception("No url specified for fetching")
            
            u = urlopen(url)
            
            path = tempfile.mkstemp()[1]
            with open(path, "wb") as f:
                meta = u.info()

                if sys.version_info >= (3, 0):
                    fileSize = int(u.getheader("Content-Length"))
                else:
                    fileSize = int(meta.getheaders("Content-Length")[0])

                fileSizeDl = 0
                blockSz = 8192
                lastEmit = 0
                while True:
                    buf = u.read(blockSz)

                    if not buf:
                        break

                    fileSizeDl += len(buf)
                    f.write(buf)
                    status = (fileSizeDl * 100 / fileSize)

                    if self.progressCallback and lastEmit != status:
                        self.progressCallback(int(status))
                        lastEmit = status;

            if self.successCallback:
                self.successCallback(path)
        except:
            logger.exception("Error occured at transfer")
            if self.errorCallback:
                self.errorCallback();


from .iq import IqProtocolEntity
from .iq_result import ResultIqProtocolEntity
from .iq_ping import PingIqProtocolEntity
from .iq_result_pong import PongResultIqProtocolEntity
from .iq_error import ErrorIqProtocolEntity
from .iq_push import PushIqProtocolEntity
from .iq_props import PropsIqProtocolEntity
from .iq_crypto import CryptoIqProtocolEntity



from yowsup.stacks import YowStack
from .layer import SyncLayer
from yowsup.layers import YowLayerEvent
from yowsup.layers.auth                        import YowCryptLayer, YowAuthenticationProtocolLayer, AuthError
from yowsup.layers.coder                       import YowCoderLayer
from yowsup.layers.network                     import YowNetworkLayer
from yowsup.layers.stanzaregulator             import YowStanzaRegulator
from yowsup.layers.protocol_receipts           import YowReceiptProtocolLayer
from yowsup.layers.protocol_acks               import YowAckProtocolLayer
from yowsup.layers.logger                      import YowLoggerLayer
from yowsup.layers.protocol_contacts           import YowContactsIqProtocolLayer
from yowsup.layers                             import YowParallelLayer

class YowsupSyncStack(object):
    def __init__(self, credentials, contacts, encryptionEnabled = False):
        """
        :param credentials:
        :param contacts: list of [jid ]
        :param encryptionEnabled:
        :return:
        """
        if encryptionEnabled:
            from yowsup.layers.axolotl                     import YowAxolotlLayer
            layers = (
                SyncLayer,
                YowParallelLayer([YowAuthenticationProtocolLayer, YowContactsIqProtocolLayer, YowReceiptProtocolLayer, YowAckProtocolLayer]),
                YowAxolotlLayer,
                YowLoggerLayer,
                YowCoderLayer,
                YowCryptLayer,
                YowStanzaRegulator,
                YowNetworkLayer
            )
        else:
            layers = (
                SyncLayer,
                YowParallelLayer([YowAuthenticationProtocolLayer, YowContactsIqProtocolLayer, YowReceiptProtocolLayer, YowAckProtocolLayer]),
                YowLoggerLayer,
                YowCoderLayer,
                YowCryptLayer,
                YowStanzaRegulator,
                YowNetworkLayer
            )

        self.stack = YowStack(layers)
        self.stack.setProp(SyncLayer.PROP_CONTACTS, contacts)
        self.stack.setProp(YowAuthenticationProtocolLayer.PROP_PASSIVE, True)
        self.stack.setCredentials(credentials)

    def start(self):
        self.stack.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECT))
        try:
            self.stack.loop()
        except AuthError as e:
            print("Authentication Error: %s" % e.message)



from .chatstate import ChatstateProtocolEntity
class IncomingChatstateProtocolEntity(ChatstateProtocolEntity):
    '''
    INCOMING

    <chatstate from="xxxxxxxxxxx@s.whatsapp.net">
    <{{composing|paused}}></{{composing|paused}}>
    </chatstate>

    OUTGOING

    <chatstate to="xxxxxxxxxxx@s.whatsapp.net">
    <{{composing|paused}}></{{composing|paused}}>
    </chatstate>
    '''

    def __init__(self, _state, _from):
        super(IncomingChatstateProtocolEntity, self).__init__(_state)
        self.setIncomingData(_from)

    def setIncomingData(self, _from):
        self._from = _from
    
    def toProtocolTreeNode(self):
        node = super(IncomingChatstateProtocolEntity, self).toProtocolTreeNode()
        node.setAttribute("from", self._from)
        return node

    def __str__(self):
        out  = super(IncomingChatstateProtocolEntity, self).__str__()
        out += "From: %s\n" % self._from
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = ChatstateProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = IncomingChatstateProtocolEntity
        entity.setIncomingData(
            node.getAttributeValue("from"),
        )
        return entity



from yowsup.layers import YowLayerEvent, YowLayerTest
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.auth import YowCryptLayer
from yowsup.layers.auth.keystream import KeyStream

class CryptLayerTest(YowLayerTest, YowCryptLayer):
    def setUp(self):
        YowCryptLayer.__init__(self)
        self.password = bytearray(list(map(ord,"password")))
        self.salt     = bytearray(list(map(ord,"salt")))
        self.username = "username"
        self.inputMessage = bytearray([0,0,0,0, ord('t'), ord('a'), ord('r'), ord('e'), ord('k')])
        self.inputMessageCrypted = bytearray(b'\x80\x00\r)\x99\xbe_\xee\x98\xecV<\x9d\x0c\xb7r')
        keys = [
            bytearray(b'>\xd5\x8a\xecB\xdb\xc8\xd4}\x98\x9aBa\x89\x9fC\x08\xdcp\x8d'),
            bytearray(b'\xd3;\xda:\x8f\x94CX\xe4;\xbb\xcc"W\x83\xe0m\xba\xe0\xd1'),
            bytearray(b'\x92w{\xc2\x04~\x08;\x81w\xb7h3\xb8\xc4t\xbd\xed\xf7q'),
            bytearray(b'.\xc7\xe4\x15\x1a\xa0\xfd\x98\xc0\xea\xefs{\r7E\xa6\xec\xd5\xfb')
        ]
        inputKey = KeyStream(keys[2], keys[3])
        outputKey = KeyStream(keys[0], keys[1])

        self.onEvent(YowLayerEvent(YowCryptLayer.EVENT_KEYS_READY, keys = (inputKey, outputKey)))

    def test_00send(self):
        self.send(self.inputMessage)
        self.assertEqual(self.lowerSink.pop(), self.inputMessageCrypted)




from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class PrivacyListIqProtocolEntity(IqProtocolEntity):
    def __init__(self, name = "default"):
        super(PrivacyListIqProtocolEntity, self).__init__("jabber:iq:privacy", _type="get")
        self.setListName(name)

    def setListName(self, name):
        self.listName = name

    def toProtocolTreeNode(self):
        node = super(PrivacyListIqProtocolEntity, self).toProtocolTreeNode()
        queryNode = ProtocolTreeNode("query")
        listNode = ProtocolTreeNode("list", {"name": self.listName})
        queryNode.addChild(listNode)
        node.addChild(queryNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = PrivacyListIqProtocolEntity
        entity.setListName(node.getChild("query").getChild("list")["name"])
        return entity



from .iq_sync import SyncIqProtocolEntity
from .iq_sync_get import GetSyncIqProtocolEntity
from .iq_sync_result import ResultSyncIqProtocolEntity
from .notification_contact_add import AddContactNotificationProtocolEntity
from .notification_contact_remove import RemoveContactNotificationProtocolEntity
from .notification_contact_update import UpdateContactNotificationProtocolEntity
from .notificiation_contacts_sync import ContactsSyncNotificationProtocolEntity
from .iq_statuses_get import GetStatusesIqProtocolEntity
from .iq_statuses_result import ResultStatusesIqProtocolEntity



from .presence              import PresenceProtocolEntity
from .presence_available    import AvailablePresenceProtocolEntity
from .presence_unavailable  import UnavailablePresenceProtocolEntity
from .presence_subscribe    import SubscribePresenceProtocolEntity
from .presence_unsubscribe  import UnsubscribePresenceProtocolEntity
from .iq_lastseen           import LastseenIqProtocolEntity
from .iq_lastseen_result    import ResultLastseenIqProtocolEntity



from axolotl.state.sessionstore import SessionStore
from axolotl.state.sessionrecord import SessionRecord
class LiteSessionStore(SessionStore):
    def __init__(self, dbConn):
        """
        :type dbConn: Connection
        """
        self.dbConn = dbConn
        dbConn.execute("CREATE TABLE IF NOT EXISTS sessions (_id INTEGER PRIMARY KEY AUTOINCREMENT,"
                       "recipient_id INTEGER UNIQUE, device_id INTEGER, record BLOB, timestamp INTEGER);")


    def loadSession(self, recipientId, deviceId):
        q = "SELECT record FROM sessions WHERE recipient_id = ? AND device_id = ?"
        c = self.dbConn.cursor()
        c.execute(q, (recipientId, deviceId))
        result = c.fetchone()

        if result:
            return SessionRecord(serialized=result[0])
        else:
            return SessionRecord()

    def getSubDeviceSessions(self, recipientId):
        q = "SELECT device_id from sessions WHERE recipient_id = ?"
        c = self.dbConn.cursor()
        c.execute(q, (recipientId,))
        result = c.fetchall()

        deviceIds = [r[0] for r in result]
        return deviceIds

    def storeSession(self, recipientId, deviceId, sessionRecord):
        self.deleteSession(recipientId, deviceId)

        q = "INSERT INTO sessions(recipient_id, device_id, record) VALUES(?,?,?)"
        c = self.dbConn.cursor()
        c.execute(q, (recipientId, deviceId, sessionRecord.serialize()))
        self.dbConn.commit()

    def containsSession(self, recipientId, deviceId):
        q = "SELECT record FROM sessions WHERE recipient_id = ? AND device_id = ?"
        c = self.dbConn.cursor()
        c.execute(q, (recipientId, deviceId))
        result = c.fetchone()

        return result is not None

    def deleteSession(self, recipientId, deviceId):
        q = "DELETE FROM sessions WHERE recipient_id = ? AND device_id = ?"
        self.dbConn.cursor().execute(q, (recipientId, deviceId))
        self.dbConn.commit()

    def deleteAllSessions(self, recipientId):
        q = "DELETE FROM sessions WHERE recipient_id = ?"
        self.dbConn.cursor().execute(q, (recipientId,))
        self.dbConn.commit()



from yowsup.layers.protocol_notifications.protocolentities.test_notification import NotificationProtocolEntityTest
from yowsup.layers.axolotl.protocolentities import EncryptNotification
from yowsup.structs import ProtocolTreeNode
class TestEncryptNotification(NotificationProtocolEntityTest):
    def setUp(self):
        super(TestEncryptNotification, self).setUp()
        self.ProtocolEntity = EncryptNotification
        self.node.addChild(ProtocolTreeNode("count", {"value": "9"}))


from yowsup.layers.protocol_acks.protocolentities.ack_incoming import IncomingAckProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest
import time

entity = IncomingAckProtocolEntity("12345", "message", "sender@s.whatsapp.com", int(time.time()))

class IncomingAckProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = IncomingAckProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.layers.protocol_contacts.protocolentities import AddContactNotificationProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import time
import unittest

entity = AddContactNotificationProtocolEntity("1234", "jid@s.whatsapp.net", int(time.time()), "notify", False,
                                                      "sender@s.whatsapp.net")

class AddContactNotificationProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        super(AddContactNotificationProtocolEntityTest, self).setUp()
        self.ProtocolEntity = AddContactNotificationProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.layers import YowProtocolLayer
from .protocolentities import * 
class YowAckProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "ack": (self.recvAckNode, self.sendAckEntity)
        }
        super(YowAckProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Ack Layer"

    def sendAckEntity(self, entity):
        self.entityToLower(entity)

    def recvAckNode(self, node):
        self.toUpper(IncomingAckProtocolEntity.fromProtocolTreeNode(node))



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_groups_participants  import ParticipantsGroupsIqProtocolEntity

class PromoteParticipantsIqProtocolEntity(ParticipantsGroupsIqProtocolEntity):
    '''
    <iq type="set" id="{{id}}" xmlns="w:g2", to="{{group_jid}}">
        <promote>
            <participant jid="{{jid}}"></participant>
            <participant jid="{{jid}}"></participant>
        </promote>
    </iq>
    '''

    def __init__(self, group_jid, participantList, _id = None):
        super(PromoteParticipantsIqProtocolEntity, self).__init__(group_jid, participantList, "promote", _id = _id)
    
    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(PromoteParticipantsIqProtocolEntity, PromoteParticipantsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = PromoteParticipantsIqProtocolEntity
        participantList = []
        for participantNode in node.getChild("promote").getAllChildren():
            participantList.append(participantNode["jid"])
        entity.setProps(node.getAttributeValue("to"), participantList)
        return entity



from axolotl.state.signedprekeystore import SignedPreKeyStore
from axolotl.state.signedprekeyrecord import SignedPreKeyRecord
from axolotl.invalidkeyidexception import InvalidKeyIdException
class LiteSignedPreKeyStore(SignedPreKeyStore):
    def __init__(self, dbConn):
        """
        :type dbConn: Connection
        """
        self.dbConn = dbConn
        dbConn.execute("CREATE TABLE IF NOT EXISTS signed_prekeys (_id INTEGER PRIMARY KEY AUTOINCREMENT,"
                       "prekey_id INTEGER UNIQUE, timestamp INTEGER, record BLOB);")


    def loadSignedPreKey(self, signedPreKeyId):
        q = "SELECT record FROM signed_prekeys WHERE prekey_id = ?"

        cursor = self.dbConn.cursor()
        cursor.execute(q, (signedPreKeyId,))

        result = cursor.fetchone()
        if not result:
            raise InvalidKeyIdException("No such signedprekeyrecord! %s " % signedPreKeyId)

        return SignedPreKeyRecord(serialized=result[0])

    def loadSignedPreKeys(self):
        q = "SELECT record FROM signed_prekeys"

        cursor = self.dbConn.cursor()
        cursor.execute(q,)
        result = cursor.fetchall()
        results = []
        for row in result:
            results.append(SignedPreKeyRecord(serialized=row[0]))

        return results

    def storeSignedPreKey(self, signedPreKeyId, signedPreKeyRecord):
        #q = "DELETE FROM signed_prekeys WHERE prekey_id = ?"
        #self.dbConn.cursor().execute(q, (signedPreKeyId,))
        #self.dbConn.commit()

        q = "INSERT INTO signed_prekeys (prekey_id, record) VALUES(?,?)"
        cursor = self.dbConn.cursor()
        cursor.execute(q, (signedPreKeyId, signedPreKeyRecord.serialize()))
        self.dbConn.commit()

    def containsSignedPreKey(self, signedPreKeyId):
        q = "SELECT record FROM signed_prekeys WHERE prekey_id = ?"
        cursor = self.dbConn.cursor()
        cursor.execute(q, (signedPreKeyId,))
        return cursor.fetchone() is not None

    def removeSignedPreKey(self, signedPreKeyId):
        q = "DELETE FROM signed_prekeys WHERE prekey_id = ?"
        cursor = self.dbConn.cursor()
        cursor.execute(q, (signedPreKeyId,))
        self.dbConn.commit()



from yowsup.layers.protocol_messages.protocolentities.message import MessageProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class MessageProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = MessageProtocolEntity
        # ORDER_MATTERS for node.toString() to output return attribs in same order
        attribs = {
            "type": "message_type",
            "id": "message-id",
            "t": "12345",
            "offline": "0",
            "from": "from_jid",
            "notify": "notify_name"
        }
        self.node = ProtocolTreeNode("message", attribs)




from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .message import MessageProtocolEntity
class TextMessageProtocolEntity(MessageProtocolEntity):
    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}" 
        offline="{{OFFLINE}}" type="text" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">
            <body>
                {{MESSAGE_DATA}}
            </body>
    </message>
    '''
    def __init__(self, body, _id = None,  _from = None, to = None, notify = None, 
        timestamp = None, participant = None, offline = None, retry = None):
        super(TextMessageProtocolEntity, self).__init__("text",_id, _from, to, notify, timestamp, participant, offline, retry)
        self.setBody(body)

    def __str__(self):
        out  = super(TextMessageProtocolEntity, self).__str__()
        out += "Body: %s\n" % self.body
        return out

    def setBody(self, body):
        self.body = body

    def getBody(self):
        return self.body

    def toProtocolTreeNode(self):
        node = super(TextMessageProtocolEntity, self).toProtocolTreeNode()
        bodyNode = ProtocolTreeNode("body", {}, None, self.body)
        node.addChild(bodyNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = MessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = TextMessageProtocolEntity
        entity.setBody(node.getChild("body").getData())
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .receipt import ReceiptProtocolEntity
from yowsup.layers.protocol_acks.protocolentities  import OutgoingAckProtocolEntity

class IncomingReceiptProtocolEntity(ReceiptProtocolEntity):

    '''
    delivered:
    <receipt to="xxxxxxxxxxx@s.whatsapp.net" id="1415389947-15"></receipt>

    read
    <receipt to="xxxxxxxxxxx@s.whatsapp.net" id="1415389947-15" type="read"></receipt>

    delivered to participant in group:
    <receipt participant="xxxxxxxxxx@s.whatsapp.net" from="yyyyyyyyyyyyy@g.us" id="1431204051-9" t="1431204094"></receipt>

    read by participant in group:
    <receipt participant="xxxxxxxxxx@s.whatsapp.net" t="1431204235" from="yyyyyyyyyyyyy@g.us" id="1431204051-9" type="read"></receipt>

    multiple items:
    <receipt type="read" from="xxxxxxxxxxxx@s.whatsapp.net" id="1431364583-191" t="1431365553">
        <list>
            <item id="1431364572-189"></item>
            <item id="1431364575-190"></item>
        </list>
    </receipt>

    multiple items to group:
    <receipt participant="xxxxxxxxxxxx@s.whatsapp.net" t="1431330533" from="yyyyyyyyyyyyyy@g.us" id="1431330385-323" type="read">
        <list>
            <item id="1431330096-317"></item>
            <item id="1431330373-320"></item>
            <item id="1431330373-321"></item>
            <item id="1431330385-322"></item>
        </list>
    </receipt>

    INCOMING
    <receipt offline="0" from="xxxxxxxxxx@s.whatsapp.net" id="1415577964-1" t="1415578027"></receipt>
    '''

    def __init__(self, _id, _from, timestamp, offline = None, type = None, participant = None, items = None):
        super(IncomingReceiptProtocolEntity, self).__init__(_id)
        self.setIncomingData(_from, timestamp, offline, type, participant, items)

    def getType(self):
        return self.type

    def getParticipant(self):
        return self.participant

    def getFrom(self, full = True):
        return self._from if full else self._from.split('@')[0]

    def setIncomingData(self, _from, timestamp, offline, type = None, participant = None, items = None):
        self._from = _from
        self.timestamp = timestamp
        self.type = type
        self.participant = participant
        if offline is not None:
            self.offline = True if offline == "1" else False
        else:
            self.offline = None
        self.items = items

    def toProtocolTreeNode(self):
        node = super(IncomingReceiptProtocolEntity, self).toProtocolTreeNode()
        node.setAttribute("from", self._from)
        node.setAttribute("t", str(self.timestamp))
        if self.offline is not None:
            node.setAttribute("offline", "1" if self.offline else "0")
        if self.type is not None:
            node.setAttribute("type", self.type)
        if self.participant is not None:
            node.setAttribute("participant", self.participant)

        if self.items is not None:
            inodes = []
            for item in self.items:
                inode = ProtocolTreeNode("item", {"id": item})
                inodes.append(inode)

            lnode = ProtocolTreeNode("list")
            lnode.addChildren(inodes)
            node.addChild(lnode)
        return node

    def __str__(self):
        out = super(IncomingReceiptProtocolEntity, self).__str__()
        out += "From: %s\n" % self._from
        out += "Timestamp: %s\n" % self.timestamp
        if self.offline is not None:
            out += "Offline: %s\n" % ("1" if self.offline else "0")
        if self.type is not None:
            out += "Type: %s\n" % (self.type)
        if self.participant is not None:
            out += "Participant: %s\n" % (self.participant)
        if self.items is not None:
            out += "Items: %s\n" % " ".join(self.items)
        return out

    def ack(self):
        return OutgoingAckProtocolEntity(self.getId(), "receipt", self.getType(), self.getFrom())

    @staticmethod
    def fromProtocolTreeNode(node):
        items = None
        listNode = node.getChild("list")
        if listNode is not None:
            items = []
            for inode in listNode.getAllChildren("item"):
               items.append(inode["id"])
        return IncomingReceiptProtocolEntity(
            node.getAttributeValue("id"),
            node.getAttributeValue("from"),
            node.getAttributeValue("t"),
            node.getAttributeValue("offline"),
            node.getAttributeValue("type"),
            node.getAttributeValue("participant"),
            items
            )



from yowsup.layers.protocol_media.protocolentities.message_media import MediaMessageProtocolEntity
from yowsup.layers.protocol_messages.protocolentities.test_message import MessageProtocolEntityTest
from yowsup.structs import ProtocolTreeNode

class MediaMessageProtocolEntityTest(MessageProtocolEntityTest):
    def setUp(self):
        super(MediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = MediaMessageProtocolEntity
        mediaNode = ProtocolTreeNode("media", {"type":"MEDIA_TYPE"}, None, None)
        self.node.addChild(mediaNode)



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_groups import GroupsIqProtocolEntity
class SubjectGroupsIqProtocolEntity(GroupsIqProtocolEntity):
    '''
    <iq type="set" id="{{id}}" xmlns="w:g2", to={{group_jid}}">
        <subject>
              {{NEW_VAL}}
        </subject>
    </iq>
    '''
    def __init__(self, jid, subject, _id = None):
        super(SubjectGroupsIqProtocolEntity, self).__init__(to = jid, _id = _id, _type = "set")
        self.setProps(subject)

    def setProps(self, subject):
        self.subject = subject

    def toProtocolTreeNode(self):
        node = super(SubjectGroupsIqProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode("subject",{}, None, self.subject))
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(SubjectGroupsIqProtocolEntity, SubjectGroupsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = SubjectGroupsIqProtocolEntity
        entity.setProps(node.getChild("subject").getData())
        return entity



from yowsup.layers import YowLayer, EventCallback
from yowsup.layers.network import YowNetworkLayer
class YowCryptLayer(YowLayer):
    '''
        send:       bytearray -> bytearray
        receive:    bytearray -> bytearray
    '''

    EVENT_KEYS_READY = "org.openwhatsapp.yowsup.crypt.keys"

    def __init__(self):
        super(YowCryptLayer, self).__init__()
        self.keys = (None,None)

    @EventCallback(YowNetworkLayer.EVENT_STATE_CONNECTED)
    def onConnected(self, yowLayerEvent):
        self.keys = (None,None)
    
    @EventCallback(EVENT_KEYS_READY)
    def onKeysReady(self, yowLayerEvent):
        self.keys = yowLayerEvent.getArg("keys")
        return True

    def send(self, data):
        outputKey = self.keys[1]
        length1 = len(data)
        if length1 > 1:
            if outputKey:
                length1 += 4
                buf = outputKey.encodeMessage(data, len(data), 0, len(data))
                res = [0,0,0]
                res.extend(buf)
                res[0] = ((8 << 4) | (length1 & 16711680) >> 16) % 256
                res[1] = ((length1 & 65280) >> 8) % 256
                res[2] = (length1 & 255) % 256
                

                data = res

            else:
                prep = [0,0,0]
                prep.extend(data)
                prep[0] = ((0 << 4) | (length1 & 16711680) >> 16) % 256
                prep[1] = ((length1 & 65280) >> 8) % 256
                prep[2] = (length1 & 255) % 256
                data = prep

        self.toLower(bytearray(data))

    def receive(self, data):
        inputKey = self.keys[0]
        metaData = data[:3]
        payload = bytearray(data[3:])

        firstByte = metaData[0]
        stanzaFlag = (firstByte & 0xF0) >> 4
        stanzaSize =  ((metaData[1] << 8) + metaData[2])  | ((firstByte & 0x0F) << 16)
        isEncrypted = ((stanzaFlag & 8) != 0)


        if inputKey and isEncrypted:
            toDecode = data[3:]
            payload = inputKey.decodeMessage(payload, 0, 4, len(payload) - 4)

        self.toUpper(payload)

    def __str__(self):
        return "Crypt Layer"


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .message_media_downloadable import DownloadableMediaMessageProtocolEntity
from yowsup.common.tools import VideoTools

class VideoDownloadableMediaMessageProtocolEntity(DownloadableMediaMessageProtocolEntity):
    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}"
        offline="{{OFFLINE}}" type="media" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">
        <media type="{{DOWNLOADABLE_MEDIA_TYPE: (image | audio | video)}}"
            mimetype="{{MIME_TYPE}}"
            filehash="{{FILE_HASH}}"
            url="{{DOWNLOAD_URL}}"
            ip="{{IP}}"
            size="{{MEDIA SIZE}}"
            file="{{FILENAME}}"


            encoding="{{ENCODING}}"
            height="{{IMAGE_HEIGHT}}"
            width="{{IMAGE_WIDTH}}"

            origin="forward"
            > {{THUMBNAIL_RAWDATA (JPEG?)}}
        </media>
    </message>
    '''
    def __init__(self,
            mimeType, fileHash, url, ip, size, fileName,
            abitrate, acodec, asampfmt, asampfreq, duration, encoding, fps,
            width, height, seconds, vbitrate, vcodec, caption = None,
            _id = None, _from = None, to = None, notify = None, timestamp = None,
            participant = None, preview = None, offline = None, retry = None):

        super(VideoDownloadableMediaMessageProtocolEntity, self).__init__("video",
            mimeType, fileHash, url, ip, size, fileName,
            _id, _from, to, notify, timestamp, participant, preview, offline, retry)
        self.setVideoProps(encoding, width, height, vbitrate, abitrate, acodec, asampfmt, asampfreq, duration, fps, seconds, vcodec, caption)

    def __str__(self):
        out  = super(VideoDownloadableMediaMessageProtocolEntity, self).__str__()
        out += "Audio bitrate: %s\n" % self.abitrate
        out += "Audio codec: %s\n" % self.acodec
        out += "Audio sampling fmt.: %s\n" % self.asampfmt
        out += "Audio sampling freq.: %s\n" % self.asampfreq
        out += "Duration: %s\n" % self.duration
        out += "Encoding: %s\n" % self.encoding
        out += "Fps: %s\n" % self.fps
        out += "Width: %s\n" % self.width
        out += "Height: %s\n" % self.height
        out += "Video bitrate: %s\n" % self.vbitrate
        out += "Video codec: %s\n" % self.vcodec
        if self.caption is not None:
            out += "Caption: %s\n" % self.caption
        return out

    def setVideoProps(self, encoding, width, height, vbitrate = None, abitrate = None, acodec = None, asampfmt = None, asampfreq = None, duration = None, fps = None, seconds = None, vcodec = None, caption = None,):
        self.abitrate  = abitrate
        self.acodec    = acodec
        self.asampfmt  = asampfmt
        self.asampfreq = asampfreq
        self.duration  = duration
        self.encoding  = encoding
        self.fps       = fps
        self.height    = height
        self.seconds   = seconds
        self.vbitrate  = vbitrate
        self.vcodec    = vcodec
        self.width     = width
        self.caption   = caption

    def getCaption(self):
        return self.caption

    def toProtocolTreeNode(self):
        node = super(VideoDownloadableMediaMessageProtocolEntity, self).toProtocolTreeNode()
        mediaNode = node.getChild("media")

        mediaNode.setAttribute("encoding",  self.encoding)
        mediaNode.setAttribute("height",    str(self.height))
        mediaNode.setAttribute("width",     str(self.width))
        if self.abitrate is not None:
        	mediaNode.setAttribute("abitrate",  str(self.abitrate))
        if self.acodec is not None:
        	mediaNode.setAttribute("acodec",    self.acodec)
        if self.asampfmt is not None:
        	mediaNode.setAttribute("asampfmt",  self.asampfmt)
        if self.asampfreq is not None:
        	mediaNode.setAttribute("asampfreq", str(self.asampfreq))
        if self.duration is not None:
        	mediaNode.setAttribute("duration",  str(self.duration))
        if self.fps is not None:
        	mediaNode.setAttribute("fps",       str(self.fps))
        if self.seconds is not None:
        	mediaNode.setAttribute("seconds",   str(self.seconds))
        if self.vbitrate is not None:
        	mediaNode.setAttribute("vbitrate",  str(self.vbitrate))
        if self.vcodec is not None:
        	mediaNode.setAttribute("vcodec",    self.vcodec)
        if self.caption is not None:
        	mediaNode.setAttribute("caption",   self.caption)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = DownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = VideoDownloadableMediaMessageProtocolEntity
        mediaNode = node.getChild("media")
        entity.setVideoProps(
        	mediaNode.getAttributeValue("encoding"),
        	mediaNode.getAttributeValue("width"),
        	mediaNode.getAttributeValue("height"),
        	mediaNode.getAttributeValue("vbitrate"),
        	mediaNode.getAttributeValue("abitrate"),
        	mediaNode.getAttributeValue("acodec"),
        	mediaNode.getAttributeValue("asampfmt"),
        	mediaNode.getAttributeValue("asampfreq"),
        	mediaNode.getAttributeValue("duration"),
        	mediaNode.getAttributeValue("fps"),
        	mediaNode.getAttributeValue("seconds"),
        	mediaNode.getAttributeValue("vcodec"),
        	mediaNode.getAttributeValue("caption")
        )
        return entity

    @staticmethod
    def fromFilePath(path, url, ip, to, mimeType = None, caption = None):
        preview = VideoTools.generatePreviewFromVideo(path)
        entity = DownloadableMediaMessageProtocolEntity.fromFilePath(path, url, DownloadableMediaMessageProtocolEntity.MEDIA_TYPE_VIDEO, ip, to, mimeType, preview)
        entity.__class__ = VideoDownloadableMediaMessageProtocolEntity

        width, height, bitrate, duration = VideoTools.getVideoProperties(path)
        assert width, "Could not determine video properties"

        duration = int(duration)
        entity.setVideoProps('raw', width, height, duration=duration, seconds=duration, caption=caption)
        return entity



from yowsup.layers.protocol_ib.protocolentities.test_ib import IbProtocolEntityTest
from yowsup.layers.protocol_ib.protocolentities.dirty_ib import DirtyIbProtocolEntity
from yowsup.structs import ProtocolTreeNode
class DirtyIbProtocolEntityTest(IbProtocolEntityTest):
    def setUp(self):
        super(DirtyIbProtocolEntityTest, self).setUp()
        self.ProtocolEntity = DirtyIbProtocolEntity
        dirtyNode = ProtocolTreeNode("dirty")
        dirtyNode["timestamp"] = "123456"
        dirtyNode["type"] = "groups"
        self.node.addChild(dirtyNode)


from yowsup.layers.protocol_presence.protocolentities.presence_unavailable import UnavailablePresenceProtocolEntity
from yowsup.layers.protocol_presence.protocolentities.test_presence import PresenceProtocolEntityTest

class UnavailablePresenceProtocolEntityTest(PresenceProtocolEntityTest):
    def setUp(self):
        super(UnavailablePresenceProtocolEntityTest, self).setUp()
        self.ProtocolEntity = UnavailablePresenceProtocolEntity
        self.node.setAttribute("type", "unavailable")



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .notification import NotificationProtocolEntity
class StatusNotificationProtocolEntity(NotificationProtocolEntity):
    '''
    <notification offline="0" id="{{NOTIFICATION_ID}}" notify="{{NOTIFY_NAME}}" type="status" 
            t="{{TIMESTAMP}}" from="{{SENDER_JID}}">
        <set>
            {{STATUS}}
        </set>
    </notification>
    
    '''

    def __init__(self, _type, _id,  _from, status, timestamp, notify, offline = False):
        super(StatusNotificationProtocolEntity, self).__init__("status", _id, _from, timestamp, notify, offline)
        self.setStatus(status)

    def setStatus(self, status):
        self.status = status
    
    def toProtocolTreeNode(self):
        node = super(StatusNotificationProtocolEntity, self).toProtocolTreeNode()
        setNode = ProtocolTreeNode("set", {}, None, self.status)
        node.addChild(setNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = NotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = StatusNotificationProtocolEntity
        entity.setStatus(node.getChild("set").getData())
        return entity


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .notification_picture import PictureNotificationProtocolEntity
class DeletePictureNotificationProtocolEntity(PictureNotificationProtocolEntity):
    '''
    <notification offline="0" id="{{NOTIFICATION_ID}}" notify="{{NOTIFY_NAME}}" type="picture" 
            t="{{TIMESTAMP}}" from="{{SENDER_JID}}">
        <delete jid="{{DELETE_JID}}">
        </delete>
    </notification>
    '''

    def __init__(self, _id,  _from, status, timestamp, notify, offline, deleteJid):
        super(DeletePictureNotificationProtocolEntity, self).__init__(_id, _from, timestamp, notify, offline)
        self.setData(deleteJid)

    def setData(self, deleteJid):
        self.deleteJid =   deleteJid

    def __str__(self):
        out = super(DeletePictureNotificationProtocolEntity, self).__str__()
        out += "Type: Delete"
        return out
    
    def toProtocolTreeNode(self):
        node = super(DeletePictureNotificationProtocolEntity, self).toProtocolTreeNode()
        deleteNode = ProtocolTreeNode("delete", {"jid": self.deleteJid}, None, None)
        node.addChild(deleteNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = PictureNotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = DeletePictureNotificationProtocolEntity
        deleteNode = node.getChild("delete")
        entity.setData(deleteNode.getAttributeValue("jid"))
        return entity



from yowsup.layers.auth.protocolentities.failure import FailureProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class FailureProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = FailureProtocolEntity
        self.node = ProtocolTreeNode("failure", {}, [ProtocolTreeNode("not-authorized", {})])


#!/usr/bin/env python
from __future__ import print_function
from setuptools import setup, find_packages
import yowsup
import platform
import sys

deps = ['python-dateutil', 'argparse', 'python-axolotl>=0.1.7']

if sys.version_info < (2,7):
    deps += ['importlib']

if platform.system().lower() == "windows":
    deps.append('pyreadline')
else:
    try:
        import readline
    except ImportError:
        deps.append('readline')

setup(
    name='yowsup2',
    version=yowsup.__version__,
    url='http://github.com/tgalal/yowsup/',
    license='GPL-3+',
    author='Tarek Galal',
    tests_require=[],
    install_requires = deps,
    scripts = ['yowsup-cli'],
    #cmdclass={'test': PyTest},
    author_email='tare2.galal@gmail.com',
    description='A WhatsApp python library',
    #long_description=long_description,
    packages= find_packages(),
    include_package_data=True,
    platforms='any',
    #test_suite='',
    classifiers = [
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        #'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ],
    #extras_require={
    #    'testing': ['pytest'],
    #}
)



from yowsup.layers import YowLayerInterface
class YowNetworkLayerInterface(YowLayerInterface):
    def connect(self):
        self._layer.createConnection()

    def disconnect(self):
        self._layer.destroyConnection()

    def getStatus(self):
        return self._layer.getStatus()


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class FailureProtocolEntity(ProtocolEntity):

    def __init__(self, reason):
        super(FailureProtocolEntity, self).__init__("failure")
        self.reason = reason

    def __str__(self):
        out  = "Failure:\n"
        out += "Reason: %s\n" % self.reason
        return out

    def getReason(self):
        return self.reason

    def toProtocolTreeNode(self):
        reasonNode = ProtocolTreeNode(self.reason, {})
        return self._createProtocolTreeNode({}, children = [reasonNode])

    @staticmethod
    def fromProtocolTreeNode(node):
        return FailureProtocolEntity( node.getAllChildren()[0].tag )


import unittest
from yowsup.env import S40YowsupEnv
class S40YowsupEnvTest(unittest.TestCase):
    def test_tokengen(self):
        phone = "1234567"
        S40YowsupEnv._TOKEN_STRING  = "PdA2DJyKoUrwLw1Bg6EIhzh502dF9noR9uFCllGk1425519315543{phone}"
        env = S40YowsupEnv()
        token = env.getToken(phone)
        self.assertEqual(token, 'e84e1f1477704159efd46f6f0781dbde')



from yowsup.layers import YowLayer, YowLayerEvent, EventCallback
from yowsup.layers.network import YowNetworkLayer
from .encoder import WriteEncoder
from .decoder import ReadDecoder
from .tokendictionary import TokenDictionary
class YowCoderLayer(YowLayer):

    PROP_DOMAIN =   "org.openwhatsapp.yowsup.prop.domain"
    PROP_RESOURCE = "org.openwhatsapp.yowsup.prop.resource"

    def __init__(self):
        YowLayer.__init__(self)
        tokenDictionary = TokenDictionary()
        self.writer = WriteEncoder(tokenDictionary)
        self.reader = ReadDecoder(tokenDictionary)
    
    @EventCallback(YowNetworkLayer.EVENT_STATE_CONNECTED)
    def onConnected(self, event):
        self.writer.reset()
        self.reader.reset()
        streamStartBytes = self.writer.getStreamStartBytes(
            self.getProp(self.__class__.PROP_DOMAIN),
            self.getProp(self.__class__.PROP_RESOURCE)
        )
        for i in range(0, 4):
            self.write(streamStartBytes.pop(0))
        self.write(streamStartBytes)        

    def send(self, data):
        self.write(self.writer.protocolTreeNodeToBytes(data))

    def receive(self, data):
        node = self.reader.getProtocolTreeNode(data)
        if node:
            self.toUpper(node)

    def write(self, i):
        if(type(i) in(list, tuple)):
            self.toLower(bytearray(i))
        else:
            self.toLower(bytearray([i]))

    def __str__(self):
        return "Coder Layer"


from yowsup.common import YowConstants
from yowsup.layers.protocol_notifications.protocolentities import NotificationProtocolEntity
from yowsup.structs import ProtocolTreeNode
class EncryptNotification(NotificationProtocolEntity):
    """
    <notification t="1419824928" id="2451228097" from="s.whatsapp.net" type="encrypt">
        <count value="9">
        </count>
    </notification>
    """
    def __init__(self, count, timestamp, _id = None, notify = None, offline = None):
        super(EncryptNotification, self).__init__("encrypt", _id, YowConstants.WHATSAPP_SERVER, timestamp, notify, offline)
        self.setProps(count)

    def setProps(self, count):
        self.count = int(count)

    def getCount(self):
        return self.count

    def toProtocolTreeNode(self):
        node = super(EncryptNotification, self).toProtocolTreeNode()
        countNode = ProtocolTreeNode("count", {"value": str(self.count)})
        node.addChild(countNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = NotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = EncryptNotification
        entity.setProps(node.getChild("count")["value"])
        return entity


import unittest
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.coder.encoder import WriteEncoder
from yowsup.layers.coder.tokendictionary import TokenDictionary

class EncoderTest(unittest.TestCase):
    def setUp(self):
        self.res = []
        self.encoder = WriteEncoder(TokenDictionary())

    def test_encode(self):
        node = ProtocolTreeNode("message", {"form": "abc", "to":"xyz"}, [ProtocolTreeNode("media", {"width" : "123"}, data="123456")])
        result = self.encoder.protocolTreeNodeToBytes(node)

        self.assertTrue(result in (
           [248, 6, 89, 252, 4, 102, 111, 114, 109, 252, 3, 97, 98, 99, 165, 252, 3, 120, 121, 122, 248, 1,
                                  248, 4, 87, 236, 99, 252, 3, 49, 50, 51, 252, 6, 49, 50, 51, 52, 53, 54],
           [248, 6, 89, 165, 252, 3, 120, 121, 122, 252, 4, 102, 111, 114, 109, 252, 3, 97, 98, 99, 248, 1,
                                  248, 4, 87, 236, 99, 252, 3, 49, 50, 51, 252, 6, 49, 50, 51, 52, 53, 54])
        )




from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class ChatstateProtocolEntity(ProtocolEntity):

    '''
    INCOMING

    <chatstate from="xxxxxxxxxxx@s.whatsapp.net">
    <{{composing|paused}}></{{composing|paused}}>
    </chatstate>

    OUTGOING

    <chatstate to="xxxxxxxxxxx@s.whatsapp.net">
    <{{composing|paused}}></{{composing|paused}}>
    </chatstate>
    '''

    STATE_TYPING = "composing"
    STATE_PAUSED = "paused"
    STATES = (STATE_TYPING, STATE_PAUSED)

    def __init__(self, _state):
        super(ChatstateProtocolEntity, self).__init__("chatstate")
        assert _state in self.__class__.STATES, "Expected chat state to be in %s, got %s" % (self.__class__.STATES, _state)
        self._state = _state

    def getState(self):
        return self._state

    def toProtocolTreeNode(self):
        node = self._createProtocolTreeNode({}, None, data = None)
        node.addChild(ProtocolTreeNode(self._state))
        return node

    def __str__(self):
        out  = "CHATSTATE:\n"
        out += "State: %s\n" % self._state
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        return ChatstateProtocolEntity(
            node.getAllChildren()[0].tag,
            )



from yowsup.structs import  ProtocolTreeNode
from .iq import IqProtocolEntity
class ErrorIqProtocolEntity(IqProtocolEntity):

    '''<iq id="1417113419-0" from="{{jid}}" type="error">
    <error text="not-acceptable" code="406" backoff="3600">
    </error>
    </iq>
    '''

    def __init__(self, _id, _from, code, text, backoff= 0 ):
        super(ErrorIqProtocolEntity, self).__init__(xmlns = None, _id = _id, _type = "error", _from = _from)
        self.setErrorProps(code, text, backoff)

    def setErrorProps(self, code, text, backoff):
        self.code = code
        self.text = text
        self.backoff = int(backoff) if backoff else 0

    def toProtocolTreeNode(self):
        node = super(ErrorIqProtocolEntity, self).toProtocolTreeNode()
        errorNode = ProtocolTreeNode("error", {"text": self.text, "code": self.code})
        if self.backoff:
            errorNode.setAttribute("backoff", str(self.backoff))
        node.addChild(errorNode)
        return node

    def __str__(self):
        out = super(ErrorIqProtocolEntity, self).__str__()
        out += "Code: %s\n" % self.code
        out += "Text: %s\n" % self.text
        out += "Backoff: %s\n" % self.backoff
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ErrorIqProtocolEntity
        errorNode = node.getChild("error")
        entity.setErrorProps(errorNode.getAttributeValue("code"),
                             errorNode.getAttributeValue("text"),
                             errorNode.getAttributeValue("backoff"))
        return entity





from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from .protocolentities import TextMessageProtocolEntity
class YowMessagesProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "message": (self.recvMessageStanza, self.sendMessageEntity)
        }
        super(YowMessagesProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Messages Layer"

    def sendMessageEntity(self, entity):
        if entity.getType() == "text":
            self.entityToLower(entity)

    ###recieved node handlers handlers
    def recvMessageStanza(self, node):
        if node.getAttributeValue("type") == "text":
            entity = TextMessageProtocolEntity.fromProtocolTreeNode(node)
            self.toUpper(entity) 




from yowsup.layers.protocol_media.protocolentities.message_media_downloadable import DownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.test_message_media import MediaMessageProtocolEntityTest

class DownloadableMediaMessageProtocolEntityTest(MediaMessageProtocolEntityTest):
    def setUp(self):
        super(DownloadableMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = DownloadableMediaMessageProtocolEntity
        mediaNode = self.node.getChild("media")
        mediaNode.setAttribute("mimetype",  "MIMETYPE")
        mediaNode.setAttribute("filehash",  "FILEHASH")
        mediaNode.setAttribute("url",       "URL")
        mediaNode.setAttribute("ip",        "IP")
        mediaNode.setAttribute("size",      "123")
        mediaNode.setAttribute("file",      "FILE")



from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode

'''
<iq xmlns="privacy" type="set" id="{{IQ_ID}}">
<privacy>
<category name="status" value="none">
<category name="profile" value="none">
<category name="last" value="none">
</category>
</privacy>
</iq>
'''

class SetPrivacyIqProtocolEntity(IqProtocolEntity):
    NAMES = ["status", "profile", "last"]
    VALUES = ["all", "contacts", "none"]
    XMLNS = "privacy"

    def __init__(self, value="all", names = None):
        # names can be a string with some element in VALUES or an array with strings with elements in VALUES
        # by default, all names are used
        super(SetPrivacyIqProtocolEntity, self).__init__(self.__class__.XMLNS, _type="set")
        self.setNames(names)
        self.setValue(value)

    @staticmethod
    def checkValidNames(names):
        names = names if names else SetPrivacyIqProtocolEntity.NAMES
        if not type(names) is list:
            names = [names]

        for name in names:
            if not name in SetPrivacyIqProtocolEntity.NAMES:
                raise Exception("Name should be in: '" + "', '".join(SetPrivacyIqProtocolEntity.NAMES) + "' but is '" + name + "'")
        return names

    @staticmethod
    def checkValidValue(value):
        if not value in SetPrivacyIqProtocolEntity.VALUES:
            raise Exception("Value should be in: '" + "', '".join(SetPrivacyIqProtocolEntity.VALUES) + "' but is '" + value + "'")
        return value

    def setNames(self, names):
        self.names = SetPrivacyIqProtocolEntity.checkValidNames(names)

    def setValue(self, value):
        self.value = SetPrivacyIqProtocolEntity.checkValidValue(value)

    def toProtocolTreeNode(self):
        node = super(SetPrivacyIqProtocolEntity, self).toProtocolTreeNode()
        queryNode = ProtocolTreeNode(self.__class__.XMLNS)
        for name in self.names:
            listNode = ProtocolTreeNode("category", {"name": name, "value": self.value})
            queryNode.addChild(listNode)
        node.addChild(queryNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = SetPrivacyIqProtocolEntity
        privacyNode = node.getChild(SetPrivacyIqProtocolEntity.XMLNS)
        names = [] 
        for categoryNode in privacyNode.getAllChildren():
            names.append(categoryNode["name"])
        entity.setNames(names)
        entity.setValue("all")
        return entity



from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from .protocolentities import ImageDownloadableMediaMessageProtocolEntity
from .protocolentities import AudioDownloadableMediaMessageProtocolEntity
from .protocolentities import VideoDownloadableMediaMessageProtocolEntity
from .protocolentities import LocationMediaMessageProtocolEntity
from .protocolentities import VCardMediaMessageProtocolEntity
from .protocolentities import RequestUploadIqProtocolEntity, ResultRequestUploadIqProtocolEntity
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity, ErrorIqProtocolEntity

class YowMediaProtocolLayer(YowProtocolLayer):

    # EVENT_REQUEST_UPLOAD = "org.openwhatsapp.org.yowsup.event.protocol_media.request_upload"

    def __init__(self):
        handleMap = {
            "message": (self.recvMessageStanza, self.sendMessageEntity),
            "iq": (self.recvIq, self.sendIq)
        }
        super(YowMediaProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Media Layer"

    # def onEvent(self, yowLayerEvent):
    #     if yowLayerEvent.getArg(self.__class__.EVENT_REQUEST_UPLOAD):
    #         fpath = yowLayerEvent.getArg("file")
    #         _type = yowLayerEvent.getArg("type")
    #         assert fpath and _type, "Must specify 'file' and 'type' in EVENT_REQUEST_UPLOAD args"
    #         entity = RequestUploadIqProtocolEntity(_type, filePath=fpath)
    #         self._sendIq(entity, self.onRequestUploadSuccess, self.onRequestUploadError)


    def sendMessageEntity(self, entity):
        if entity.getType() == "media":
            self.entityToLower(entity)

    def recvMessageStanza(self, node):
        if node.getAttributeValue("type") == "media":
            mediaNode = node.getChild("media")
            if mediaNode.getAttributeValue("type") == "image":
                entity = ImageDownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("type") == "audio":
                entity = AudioDownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("type") == "video":
                entity = VideoDownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("type") == "location":
                entity = LocationMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("type") == "vcard":
                entity = VCardMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)

    def sendIq(self, entity):
        """
        :type entity: IqProtocolEntity
        """
        if entity.getType() == IqProtocolEntity.TYPE_SET and entity.getXmlns() == "w:m":
            #media upload!
            self._sendIq(entity, self.onRequestUploadSuccess, self.onRequestUploadError)

    def recvIq(self, node):
        """
        :type node: ProtocolTreeNode
        """

    def onRequestUploadSuccess(self, resultNode, requestUploadEntity):
        self.toUpper(ResultRequestUploadIqProtocolEntity.fromProtocolTreeNode(resultNode))

    def onRequestUploadError(self, errorNode, requestUploadEntity):
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(errorNode))



from yowsup.structs import ProtocolTreeNode
from .iq_sync import SyncIqProtocolEntity
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity

class ResultStatusesIqProtocolEntity(IqProtocolEntity):
    '''
    <iq type="result" from="s.whatsapp.net" id="1">
        <status>
            <user jid="{number}@s.whatsapp.net" t="1330555420">
                {status message}
                HEX:{status message in hex}
            </user>
            <user jid="{number}@s.whatsapp.net" t="1420813055">
                {status message}
                HEX:{status message in hex}
            </user>
        </status>
    </iq>
    '''
    XMLNS = 'status'
    def __init__(self, _id, _from, statuses):
        super(ResultStatusesIqProtocolEntity, self).__init__(self.__class__.XMLNS, _id, 'result', _from=_from)
        self.setResultStatusesProps(statuses)

    def setResultStatusesProps(self, statuses):
        assert type(statuses) is dict, "statuses must be dict"
        self.statuses = statuses

    def __str__(self):
        out = super(ResultStatusesIqProtocolEntity, self).__str__()
        out += "Statuses: %s\n" % ','.join(jid + '(' + str(v) + ')' for jid, v in self.statuses.items())
        return out

    def toProtocolTreeNode(self):
        node = super(ResultStatusesIqProtocolEntity, self).toProtocolTreeNode()
        users = [ProtocolTreeNode('user', {'jid': jid, 't': t}, None, status) for jid, (status, t) in self.statuses.items()]
        statusNode = ProtocolTreeNode('status', None, users)
        node.addChild(statusNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        statusNode = node.getChild('status')
        users = statusNode.getAllChildren()
        statuses = dict()
        for user in users:
            statuses[user['jid']] = (user.getData(), user['t'])

        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ResultStatusesIqProtocolEntity
        entity.setResultStatusesProps(statuses)
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class PresenceProtocolEntity(ProtocolEntity):
    '''
    <presence type="{{type}} name={{push_name}}"></presence>
    Should normally be either type or name

    when contact goes offline:
    <presence type="unavailable" from="{{contact_jid}}" last="deny | ?">
    </presence>

    when contact goes online:
    <presence from="contact_jid">
    </presence>

    '''

    def __init__(self, _type = None, name = None, _from = None, last = None):
        super(PresenceProtocolEntity, self).__init__("presence")
        self._type = _type
        self.name = name
        self._from = _from
        self.last = last

    def getType(self):
        return self._type

    def getName(self):
        return self.name

    def getFrom(self, full = True):
        return self._from if full else self._from.split('@')[0]

    def getLast(self):
        return self.last

    def toProtocolTreeNode(self):
        attribs = {}
        if self._type:
            attribs["type"] = self._type
        if self.name:
            attribs["name"] = self.name
        if self._from:
            attribs["from"] = self._from
        if self.last:
            attribs["last"] = self.last

        return self._createProtocolTreeNode(attribs, None, None)

    def __str__(self):
        out  = "Presence:\n"
        if self._type:
            out += "Type: %s\n" % self._type
        if self.name:
            out += "Name: %s\n" % self.name
        if self._from:
            out += "From: %s\n" % self._from
        if self.last:
            out += "Last seen: %s\n" % self.last
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        return PresenceProtocolEntity(
            node.getAttributeValue("type"),
            node.getAttributeValue("name"),
            node.getAttributeValue("from"),
            node.getAttributeValue("last")
            )




from axolotl.state.prekeystore import PreKeyStore
from axolotl.state.prekeyrecord import PreKeyRecord
class LitePreKeyStore(PreKeyStore):
    def __init__(self, dbConn):
        """
        :type dbConn: Connection
        """
        self.dbConn = dbConn
        dbConn.execute("CREATE TABLE IF NOT EXISTS prekeys (_id INTEGER PRIMARY KEY AUTOINCREMENT,"
                       "prekey_id INTEGER UNIQUE, sent_to_server BOOLEAN, record BLOB);")

    def loadPreKey(self, preKeyId):
        q = "SELECT record FROM prekeys WHERE prekey_id = ?"

        cursor = self.dbConn.cursor()
        cursor.execute(q, (preKeyId,))

        result = cursor.fetchone()
        if not result:
            raise Exception("No such prekeyRecord!")

        return PreKeyRecord(serialized = result[0])

    def loadPendingPreKeys(self):
        q = "SELECT record FROM prekeys"
        cursor = self.dbConn.cursor()
        cursor.execute(q)
        result = cursor.fetchall()

        return [PreKeyRecord(serialized=result[0]) for result in result]

    def storePreKey(self, preKeyId, preKeyRecord):
        #self.removePreKey(preKeyId)
        q = "INSERT INTO prekeys (prekey_id, record) VALUES(?,?)"
        cursor = self.dbConn.cursor()
        cursor.execute(q, (preKeyId, preKeyRecord.serialize()))
        self.dbConn.commit()

    def containsPreKey(self, preKeyId):
        q = "SELECT record FROM prekeys WHERE prekey_id = ?"
        cursor = self.dbConn.cursor()
        cursor.execute(q, (preKeyId,))
        return cursor.fetchone() is not None

    def removePreKey(self, preKeyId):
        q = "DELETE FROM prekeys WHERE prekey_id = ?"
        cursor = self.dbConn.cursor()
        cursor.execute(q, (preKeyId,))
        self.dbConn.commit()



class WriteEncoder:

    def __init__(self, tokenDictionary):
        self.tokenDictionary = tokenDictionary
        self.streamStarted = False

    def reset(self):
        self.streamStarted = False

    def getStreamStartBytes(self, domain, resource):
        data = []
        self.streamStarted = True
        data.append(87)
        data.append(65)
        data.append(1)
        data.append(5)

        streamOpenAttributes = {"to": domain, "resource": resource}
        self.writeListStart(len(streamOpenAttributes) * 2 + 1, data)
        data.append(1);
        self.writeAttributes(streamOpenAttributes, data)
        return data

    def protocolTreeNodeToBytes(self, node):
        outBytes = []
        self.writeInternal(node, outBytes)

        return outBytes


    def writeInternal(self, node, data):

        x = 1 + \
        (0 if node.attributes is None else len(node.attributes) * 2) + \
        (0 if not node.hasChildren() else 1) + \
        (0 if node.data is None else 1)

        self.writeListStart(x, data)

        self.writeString(node.tag, data)
        self.writeAttributes(node.attributes, data);

        if node.data is not None:
            self.writeBytes(node.data, data)

        if node.hasChildren():
            self.writeListStart(len(node.children), data);
            for c in node.children:
                self.writeInternal(c, data);


    def writeAttributes(self, attributes, data):
        if attributes is not None:
            for key, value in attributes.items():
                self.writeString(key, data);
                self.writeString(value, data);


    def writeBytes(self, bytes, data):

        length = len(bytes)
        if length >= 256:
            data.append(253)
            self.writeInt24(length, data)
        else:
            data.append(252)
            self.writeInt8(length, data)

        for b in bytes:
            if type(b) is int:
                data.append(b)
            else:
                data.append(ord(b))

    def writeInt8(self, v, data):
        data.append(v & 0xFF)


    def writeInt16(self, v, data):
        data.append((v & 0xFF00) >> 8);
        data.append((v & 0xFF) >> 0);


    def writeInt24(self, v, data):
        data.append((v & 0xFF0000) >> 16)
        data.append((v & 0xFF00) >> 8)
        data.append((v & 0xFF) >> 0)


    def writeListStart(self, i, data):
        if i == 0:
            data.append(0)
        elif i < 256:
            data.append(248)
            self.writeInt8(i, data)
        else:
            data.append(249)
            self.writeInt16(i, data)

    def writeToken(self, intValue, data):
        if intValue < 245:
            data.append(intValue)
        elif intValue <=500:
            data.append(254)
            data.append(intValue - 245)

    def writeString(self, tag, data):
        tok = self.tokenDictionary.getIndex(tag)
        if tok:
            index, secondary = tok
            if secondary:
                self.writeToken(236, data)
            self.writeToken(index, data)
        else:
            at = '@'.encode() if type(tag) == bytes else '@'
            try:
                atIndex = tag.index(at)

                if atIndex < 1:
                    raise ValueError("atIndex < 1")
                else:
                    server = tag[atIndex+1:]
                    user = tag[0:atIndex]
                    self.writeJid(user, server, data)
            except ValueError:
                self.writeBytes(self.encodeString(tag), data)

    def encodeString(self, string):
        res = []

        if type(string) == bytes:
            for char in string:
                res.append(char)
        else:
            for char in string:
                res.append(ord(char))
        return res;

    def writeJid(self, user, server, data):
        data.append(250)
        if user is not None:
            self.writeString(user, data)
        else:
            self.writeToken(0, data)
        self.writeString(server, data)



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class StreamFeaturesProtocolEntity(ProtocolEntity):
    def __init__(self,  features = None):
        super(StreamFeaturesProtocolEntity, self).__init__("stream:features")
        self.setFeatures(features)

    def setFeatures(self, features = None):
        self.features = features or []

    def toProtocolTreeNode(self):
        featureNodes = [ProtocolTreeNode(feature) for feature in self.features]
        return self._createProtocolTreeNode({}, children = featureNodes, data = None)


    @staticmethod
    def fromProtocolTreeNode(node):
        return StreamFeaturesProtocolEntity([fnode.tag for fnode in node.getAllChildren()])


from .notification_groups import GroupsNotificationProtocolEntity
from yowsup.structs import ProtocolTreeNode
class CreateGroupsNotificationProtocolEntity(GroupsNotificationProtocolEntity):
    '''
<notification notify="WhatsApp" id="{{id}}" t="1420402514" participant="{{participant_jiid}}" from="{{group_jid}}" type="w:gp2">
<create type="new">
<group subject="{{GROUPSUBJ}}" creation="{{GROUP_CREATION_TYIME}}" creator="{{CREATOR_JID}}"
    s_t="{{SUBJECT_SET_TIMESTAMP}}" id="{{GROUP_ID}}" s_o="{{SUBJECT_OWNER_JID}}">
<participant type="admin" jid="{{JID_1}}">
</participant>
<participant jid="{{JID_2}}">
</participant>
</group>
</create>
</notification>
    '''
    TYPE_CREATE_NEW = "new"
    TYPE_PARTICIPANT_ADMIN = "admin"
    def __init__(self, _id, _from, timestamp, notify, participant, offline,
                 createType, groupId, creationTimestamp, creatorJid,
                 subject, subjectTime, subjectOwnerJid,
                 participants):
        super(CreateGroupsNotificationProtocolEntity, self).__init__(_id, _from, timestamp, notify, participant, offline)
        self.setGroupProps(createType, groupId, creationTimestamp, creatorJid,
                           subject, subjectTime, subjectOwnerJid, participants)

    def setGroupProps(self, createType, groupId, creationTimestamp, creatorJid,
                      subject, subjectTime, subjectOwnerJid,
                      participants):

        assert type(participants) is dict, "Participants must be a dict {jid => type?}"

        self.createType = createType
        self.groupId = groupId
        self.creationTimestamp = int(creationTimestamp)
        self.creatorJid = creatorJid
        self.subject = subject
        self.subjectTime = int(subjectTime)
        self.subjectOwnerJid = subjectOwnerJid
        self.participants = participants

    def getParticipants(self):
        return self.participants

    def getSubject(self):
        return self.subject

    def getGroupId(self):
        return self.groupId

    def getCreationTimestamp(self):
        return self.creationTimestamp

    def getCreatorJid(self, full = True):
        return self.creatorJid if full else self.creatorJid.split('@')[0]

    def getSubjectTimestamp(self):
        return self.subjectTime

    def getSubjectOwnerJid(self, full = True):
        return self.subjectOwnerJid if full else self.subjectOwnerJid.split('@')[0]

    def getCreatetype(self):
        return self.createType

    def getGroupAdmin(self, full = True):
        for jid, _type in self.participants.items():
            if _type == self.__class__.TYPE_PARTICIPANT_ADMIN:
                return jid if full else jid.split('@')[0]

    def __str__(self):
        out = super(CreateGroupsNotificationProtocolEntity, self).__str__()
        out += "Creator: %s\n" % self.getCreatorJid()
        out += "Create type: %s\n" % self.getCreatetype()
        out += "Creation timestamp: %s\n" % self.getCreationTimestamp()
        out += "Subject: %s\n" % self.getSubject()
        out += "Subject owner: %s\n" % self.getSubjectOwnerJid()
        out += "Subject timestamp: %s\n" % self.getSubjectTimestamp()
        out += "Participants: %s\n" % self.getParticipants()
        return out

    def toProtocolTreeNode(self):
        node = super(CreateGroupsNotificationProtocolEntity, self).toProtocolTreeNode()
        createNode = ProtocolTreeNode("create", {"type": self.getCreatetype()})
        groupNode = ProtocolTreeNode("group", {
            "subject": self.getSubject(),
            "creation": str(self.getCreationTimestamp()),
            "creator": self.getCreatorJid(),
            "s_t": self.getSubjectTimestamp(),
            "s_o": self.getSubjectOwnerJid(),
            "id": self.getGroupId()
        })

        participants = []
        for jid, _type in self.getParticipants().items():
            pnode = ProtocolTreeNode("participant", {"jid": jid})
            if _type:
                pnode["type"] = _type
            participants.append(pnode)

        groupNode.addChildren(participants)
        createNode.addChild(groupNode)
        node.addChild(createNode)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        createNode = node.getChild("create")
        groupNode = createNode.getChild("group")
        participants = {}
        for p in groupNode.getAllChildren("participant"):
            participants[p["jid"]] = p["type"]

        return CreateGroupsNotificationProtocolEntity(
            node["id"], node["from"], node["t"], node["notify"], node["participant"], node["offline"],
            createNode["type"], groupNode["id"], groupNode["creation"], groupNode["creator"], groupNode["subject"],
            groupNode["s_t"], groupNode["s_o"], participants
        )



from yowsup.layers.auth.protocolentities.challenge import ChallengeProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class ChallengeProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = ChallengeProtocolEntity
        attribs             = {}
        childNodes          = []
        data                = "dummydata"
        self.node = ProtocolTreeNode("challenge", attribs, [], data)


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from yowsup.layers.protocol_messages.protocolentities import MessageProtocolEntity
class MediaMessageProtocolEntity(MessageProtocolEntity):
    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}" 
        offline="{{OFFLINE}}" type="text" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">
        <media type="{{DOWNLOADABLE_MEDIA_TYPE: (image | audio | video)}}"
            > {{THUMBNAIL_RAWDATA (JPEG?)}}
        </media>
    </message>
    '''

    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}" 
        offline="{{OFFLINE}}" type="text" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">

        <media type="audio"
            
            mimetype="audio/aac"
            filehash="86nonv++wq864nlkmbhJXZyPbILLlQ2KyYZxsLo8z1g="
            url="https://mms884.whatsapp.net/d/h72eS2EAV6YrHfZLBaqzFVRffGUABQdt_-pVhg/Arlt8j7XkRsfFw22i-KRffXxl7j9iVsYLbJN4APwsKGJ.aac"
            ip="174.37.199.214"
            size="6003"
            file="Arlt8j7XkRsfFw22i-KRffXxl7j9iVsYLbJN4APwsKGJ.aac"


            origin="live"
            seconds="1" 
            abitrate="32"
            duration="1"
            asampfreq="22050"
            acodec="aac"
        ></media>

        <media type="video"

            mimetype="video/mp4"
            filehash="b/f7d5bF7wK7GPqo4kH7qeka/JG0KzkFjUC4Veq9Iwg="
            url="https://mms880.whatsapp.net/d/Y00AgRfYWMWZQXTXgy_FJ1Rfe9YABQdt914Dvg/ArZwkA1pbhHdyC5eXRzgPb-DCK4K7PooUUi0kYGxm-wj.mp4"
            ip="173.193.205.8"
            size="112928"
            file="ArZwkA1pbhHdyC5eXRzgPb-DCK4K7PooUUi0kYGxm-wj.mp4"

            fps="25"
            encoding="raw"
            seconds="1"
            vcodec="h264"
            abitrate="60"
            vbitrate="726"
            height="360"
            asampfmt="flt"
            duration="1"
            asampfreq="44100"
            acodec="aac"
            width="480"
        >{{THUMBNAIL_RAWDATA}}</media>


        <media 
            latitude="52.52393" 
            type="location"
            longitude="13.41747"
            encoding="raw"
        >{{THUMBNAIL_RAWDATA}}</media>
            

        <media type="vcard">
            <vcard name="Hany Yasser">
                BEGIN:VCARD
                VERSION:3.0
                N:Yasser;Hany;;;
                FN:Hany Yasser
                PHOTO;BASE64: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
                BDAY;value=date:1989-01-05
                ORG:Vodafone Egypt;
                item1.EMAIL;type=INTERNET:hanyyasser@hotmail.com
                item1.X-ABLabel:INTERNET
                item2.EMAIL;type=INTERNET:hanybotbot@hotmail.com
                item2.X-ABLabel:INTERNET
                item3.ADR;type=HOME:;;Heliopolis;Cairo;Al Qahirah;;Egypt
                item4.ADR;type=HOME:;;;cairo;;;Egypt
                item5.URL:http://www.facebook.com/profile.php?id=626850952
                item5.X-ABLabel:_$!<HomePage>!$_
                X-FACEBOOK:hany.yasser1
                END:VCARD
            </vcard>

        </media>

    </message>
    '''


    MEDIA_TYPE_IMAGE = "image"
    MEDIA_TYPE_VIDEO = "video"
    MEDIA_TYPE_AUDIO = "audio"
    MEDIA_TYPE_VCARD = "vcard"
    MEDIA_TYPE_LOCATION = "location"
    TYPES_MEDIA = (MEDIA_TYPE_AUDIO, MEDIA_TYPE_IMAGE, MEDIA_TYPE_VIDEO, MEDIA_TYPE_VCARD, MEDIA_TYPE_LOCATION)

    def __init__(self, mediaType, _id = None, _from = None, to = None, notify = None, timestamp = None, participant = None, preview = None, offline = None, retry = None):
        super(MediaMessageProtocolEntity, self).__init__("media", _id, _from, to, notify, timestamp, participant, offline, retry)
        self.setMediaType(mediaType)
        self.setPreview(preview)

    def __str__(self):
        out  = super(MediaMessageProtocolEntity, self).__str__()
        out += "Media Type: %s\n" % self.mediaType
        out += "Has Preview: %s\n" % (self.preview is not None)
        return out

    def setPreview(self, preview):
        self.preview = preview

    def getPreview(self):
        return self.preview

    def setMediaType(self, mediaType):
        self.mediaType = mediaType

    def getMediaType(self):
        return self.mediaType


    def toProtocolTreeNode(self):
        node = super(MediaMessageProtocolEntity, self).toProtocolTreeNode()
        mediaNode = ProtocolTreeNode("media", {"type": self.mediaType}, None, None)
        node.addChild(mediaNode)
        if self.preview:
            mediaNode.setData(self.preview)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = MessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = MediaMessageProtocolEntity
        entity.setMediaType(node.getChild("media").getAttributeValue("type"))
        preview = node.getChild("media").getData()
        entity.setPreview(preview)
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .presence import PresenceProtocolEntity
class UnavailablePresenceProtocolEntity(PresenceProtocolEntity):
    '''
    <presence type="unavailable"></presence>
    response:
    <presence type="unavailable" from="self.jid">
    </presence>
    '''
    def __init__(self):
        super(UnavailablePresenceProtocolEntity, self).__init__("unavailable")


from yowsup.common import YowConstants
from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_groups import GroupsIqProtocolEntity
class ListGroupsIqProtocolEntity(GroupsIqProtocolEntity):
    '''
    <iq id="{{id}}"" type="get" to="g.us" xmlns="w:g2">
        <"{{participating | owning}}"></"{{participating | owning}}">
    </iq>
    
    result (processed in iq_result_groups_list.py):
    <iq type="result" from="g.us" id="{{IQ_ID}}">
      <groups>
          <group s_t="{{SUBJECT_TIME}}" creation="{{CREATING_TIME}}" creator="{{OWNER_JID}}" id="{{GROUP_ID}}" s_o="{{SUBJECT_OWNER_JID}}" subject="{{SUBJECT}}">
            <participant jid="{{JID}}" type="admin">
            </participant>
            <participant jid="{{JID}}">
            </participant>
          </group>
          <group s_t="{{SUBJECT_TIME}}" creation="{{CREATING_TIME}}" creator="{{OWNER_JID}}" id="{{GROUP_ID}}" s_o="{{SUBJECT_OWNER_JID}}" subject="{{SUBJECT}}">
            <participant jid="{{JID}}" type="admin">
            </participant>
          </group>
      <groups>
    </iq>
    '''

    GROUP_TYPE_PARTICIPATING = "participating"
    GROUP_TYPE_OWNING        = "owning"

    GROUPS_TYPES = (GROUP_TYPE_PARTICIPATING, GROUP_TYPE_OWNING)

    def __init__(self, groupsType = GROUP_TYPE_PARTICIPATING, _id = None):
        super(ListGroupsIqProtocolEntity, self).__init__(_id=_id, to = YowConstants.WHATSAPP_GROUP_SERVER, _type = "get")
        self.setProps(groupsType)

    def setProps(self, groupsType):
        assert groupsType in self.__class__.GROUPS_TYPES,\
            "Groups type must be %s, not %s" % (" or ".join(self.__class__.GROUPS_TYPES), groupsType)
        self.groupsType = groupsType

    def toProtocolTreeNode(self):
        node = super(ListGroupsIqProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode(self.groupsType))
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(ListGroupsIqProtocolEntity, ListGroupsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = ListGroupsIqProtocolEntity
        entity.setProps(node.getChild(0).tag)
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .notification_picture import PictureNotificationProtocolEntity
class SetPictureNotificationProtocolEntity(PictureNotificationProtocolEntity):
    '''
    <notification offline="0" id="{{NOTIFICATION_ID}}" notify="{{NOTIFY_NAME}}" type="picture" 
            t="{{TIMESTAMP}}" from="{{SENDER_JID}}">
        <set jid="{{SET_JID}}" id="{{SET_ID}}">
        </set>
    </notification>
    '''

    def __init__(self, _id,  _from, status, timestamp, notify, offline, setJid, setId):
        super(SetPictureNotificationProtocolEntity, self).__init__(_id, _from, timestamp, notify, offline)
        self.setData(setJid, setId)

    def setData(self, setJid, setId):
        self.setId  =    setId
        self.setJid =   setJid
    
    def toProtocolTreeNode(self):
        node = super(SetPictureNotificationProtocolEntity, self).toProtocolTreeNode()
        setNode = ProtocolTreeNode("set", {"jid": self.setJid, "id": self.setId}, None, None)
        node.addChild(setNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = PictureNotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = SetPictureNotificationProtocolEntity
        setNode = node.getChild("set")
        entity.setData(setNode.getAttributeValue("jid"), setNode.getAttributeValue("id"))
        return entity